import argparse
import typing

import numpy

from goats.core import aliased
from goats.core import iterables


_constants = {
    'pi': {
        'info': "The ratio of a circle's circumference to its diameter.",
        'all': numpy.pi,
        'quantity': 'number',
    },
    'k': {
        'info': "Boltzmann's constant.",
        'mks': {'unit': 'J / K', 'value': 1.3807e-23},
        'cgs': {'unit': 'erg / K', 'value': 1.3807e-16},
        'quantity': 'energy / temperature'
    },
    'e': {
        'info': "Elementary charge.",
        'mks': {'unit': 'C', 'value': 1.6022e-19},
        'cgs': {'unit': 'statC', 'value': 4.8032e-10},
        'quantity': 'charge',
    },
    'me': {
        'info': "Electron mass.",
        'mks': {'unit': 'kg', 'value': 9.1094e-31},
        'cgs': {'unit': 'g', 'value': 9.1094e-28},
        'quantity': 'mass',
    },
    'mp': {
        'info': "Proton mass.",
        'mks': {'unit': 'kg', 'value': 1.6726e-27},
        'cgs': {'unit': 'g', 'value': 1.6726e-24},
        'quantity': 'mass',
    },
    'G': {
        'info': "Gravitational constant.",
        'mks': {'unit': 'm^3 / (s^2 * kg)', 'value': 6.6726e-11},
        'cgs': {'unit': 'dyn * cm^2 / g^2', 'value': 6.6726e-8},
        'quantity': 'force * area / mass^2',
    },
    'g': {
        'info': "Gravitational acceleration.",
        'mks': {'unit': 'm / s^2', 'value': 9.8067},
        'cgs': {'unit': 'cm / s^2', 'value': 9.8067e2},
        'quantity': 'acceleration',
    },
    'h': {
        'info': "Planck's constant.",
        'mks': {'unit': 'J * s', 'value': 6.6261e-34},
        'cgs': {'unit': 'erg * s', 'value': 6.6261e-27},
        'quantity': 'energy * time',
    },
    'c': {
        'info': "Speed of light in a vacuum.",
        'mks': {'unit': 'm / s', 'value': 2.9979e8},
        'cgs': {'unit': 'cm / s', 'value': 2.9979e10},
        'quantity': 'speed',
    },
    'epsilon0': {
        'info': "Permittivity of free space.",
        'mks': {'unit': 'F / m', 'value': 8.8542e-12},
        'cgs': {'unit': '1', 'value': 1.0},
        'quantity': 'permittivity',
    },
    'mu0': {
        'info': "Permeability of free space.",
        'mks': {'unit': 'H / m', 'value': 4*numpy.pi * 1e-7},
        'cgs': {'unit': '1', 'value': 1.0},
        'quantity': 'permeability',
    },
    'Rinfinity': {
        'info': "Rydberg constant.",
        'mks': {'unit': '1 / m', 'value': 1.0974e7},
        'cgs': {'unit': '1 / cm', 'value': 1.0974e5},
        'quantity': '1 / length',
    },
    'a0': {
        'info': "Bohr radius.",
        'mks': {'unit': 'm', 'value': 5.2918e-11},
        'cgs': {'unit': 'cm', 'value': 5.2918e-9},
        'quantity': 'length',
    },
    're': {
        'info': "Classical electron radius.",
        'mks': {'unit': 'm', 'value': 2.8179e-15},
        'cgs': {'unit': 'cm', 'value': 2.8179e-13},
        'quantity': 'length',
    },
    'alpha': {
        'info': "Fine structure constant.",
        'mks': {'unit': '1', 'value': 7.2974e-3},
        'cgs': {'unit': '1', 'value': 7.2974e-3},
        'quantity': 'number',
    },
    'c1': {
        'info': "First radiation constant.",
        'mks': {'unit': 'W * m^2', 'value': 3.7418e-16},
        'cgs': {'unit': 'erg * cm^2 / s', 'value': 3.7418e-16},
        'quantity': 'power * area',
    },
    'c2': {
        'info': "Second radiation constant.",
        'mks': {'unit': 'm * K', 'value': 1.4388e-2},
        'cgs': {'unit': 'cm * K', 'value': 1.4388},
        'quantity': 'length * temperature',
    },
    'sigma': {
        'info': "Stefan-Boltzmann constant.",
        'mks': {'unit': 'W / (m^2 * K^4)', 'value': 5.6705e-8},
        'cgs': {'unit': '(erg / s) / (cm^2 * K^4)', 'value': 5.6705e-5},
        'quantity': 'power / (area * temperature^4)',
    },
    'eV': {
        'info': "Energy associated with 1 eV.",
        'mks': {'unit': 'J', 'value': 1.6022e-19},
        'cgs': {'unit': 'erg', 'value': 1.6022e-12},
        'quantity': 'energy',
    },
    'amu': {
        'info': "Atomic mass unit.",
        'mks': {'unit': 'kg', 'value': 1.6605e-27},
        'cgs': {'unit': 'g', 'value': 1.6605e-24},
        'quantity': 'mass',
    },
    'au': {
        'info': "Astronomical unit.",
        'mks': {'unit': 'm', 'value': 1.495978707e11},
        'cgs': {'unit': 'cm', 'value': 1.495978707e13},
        'quantity': 'length',
    },
}
_constants['H+'] = {
    'info': "First ionization energy of hydrogen.",
    **{
        k: {
            'unit': _constants['eV'][k]['unit'],
            'value': 13.6 * _constants['eV'][k]['value'],
        }
        for k in ('mks', 'cgs')
    },
    'quantity': 'energy',
}
def _normalize(defined: typing.Dict[str, dict]):
    """Normalize constant definitions."""
    systems = ('mks', 'cgs')
    norm = {}
    for key, definition in defined.items():
        if 'all' in definition:
            value = definition.pop('all')
            new = {k: {'unit': '1', 'value': value} for k in systems}
            definition.update(new)
        norm[key] = definition
    return norm
CONSTANTS = aliased.Mapping(_normalize(_constants))


_elements = [
    {'symbol': 'H', 'name': 'Hydrogen', 'mass': 1.00797},
    {'symbol': 'He', 'name': 'Helium', 'mass': 4.00260},
    {'symbol': 'Li', 'name': 'Lithium', 'mass': 6.941},
    {'symbol': 'Be', 'name': 'Beryllium', 'mass': 9.01218},
    {'symbol': 'B', 'name': 'Boron', 'mass': 10.81},
    {'symbol': 'C', 'name': 'Carbon', 'mass': 12.011},
    {'symbol': 'N', 'name': 'Nitrogen', 'mass': 14.0067},
    {'symbol': 'O', 'name': 'Oxygen', 'mass': 15.9994},
    {'symbol': 'F', 'name': 'Fluorine', 'mass': 18.998403},
    {'symbol': 'Ne', 'name': 'Neon', 'mass': 20.179},
    {'symbol': 'Na', 'name': 'Sodium', 'mass': 22.98977},
    {'symbol': 'Mg', 'name': 'Magnesium', 'mass': 24.305},
    {'symbol': 'Al', 'name': 'Aluminum', 'mass': 26.98154},
    {'symbol': 'Si', 'name': 'Silicon', 'mass': 28.0855},
    {'symbol': 'P', 'name': 'Phosphorus', 'mass': 30.97376},
    {'symbol': 'S', 'name': 'Sulfur', 'mass': 32.06},
    {'symbol': 'Cl', 'name': 'Chlorine', 'mass': 35.453},
    {'symbol': 'K', 'name': 'Potassium', 'mass': 39.0983},
    {'symbol': 'Ar', 'name': 'Argon', 'mass': 39.948},
    {'symbol': 'Ca', 'name': 'Calcium', 'mass': 40.08},
    {'symbol': 'Sc', 'name': 'Scandium', 'mass': 44.9559},
    {'symbol': 'Ti', 'name': 'Titanium', 'mass': 47.90},
    {'symbol': 'V', 'name': 'Vanadium', 'mass': 50.9415},
    {'symbol': 'Cr', 'name': 'Chromium', 'mass': 51.996},
    {'symbol': 'Mn', 'name': 'Manganese', 'mass': 54.9380},
    {'symbol': 'Fe', 'name': 'Iron', 'mass': 55.847},
    {'symbol': 'Ni', 'name': 'Nickel', 'mass': 58.70},
    {'symbol': 'Co', 'name': 'Cobalt', 'mass': 58.9332},
    {'symbol': 'Cu', 'name': 'Copper', 'mass': 63.546},
    {'symbol': 'Zn', 'name': 'Zinc', 'mass': 65.38},
    {'symbol': 'Ga', 'name': 'Gallium', 'mass': 69.72},
    {'symbol': 'Ge', 'name': 'Germanium', 'mass': 72.59},
    {'symbol': 'As', 'name': 'Arsenic', 'mass': 74.9216},
    {'symbol': 'Se', 'name': 'Selenium', 'mass': 78.96},
    {'symbol': 'Br', 'name': 'Bromine', 'mass': 79.904},
    {'symbol': 'Kr', 'name': 'Krypton', 'mass': 83.80},
    {'symbol': 'Rb', 'name': 'Rubidium', 'mass': 85.4678},
    {'symbol': 'Sr', 'name': 'Strontium', 'mass': 87.62},
    {'symbol': 'Y', 'name': 'Yttrium', 'mass': 88.9059},
    {'symbol': 'Zr', 'name': 'Zirconium', 'mass': 91.22},
    {'symbol': 'Nb', 'name': 'Niobium', 'mass': 92.9064},
    {'symbol': 'Mo', 'name': 'Molybdenum', 'mass': 95.94},
    {'symbol': 'Tc', 'name': 'Technetium', 'mass': (98)},
    {'symbol': 'Ru', 'name': 'Ruthenium', 'mass': 101.07},
    {'symbol': 'Rh', 'name': 'Rhodium', 'mass': 102.9055},
    {'symbol': 'Pd', 'name': 'Palladium', 'mass': 106.4},
    {'symbol': 'Ag', 'name': 'Silver', 'mass': 107.868},
    {'symbol': 'Cd', 'name': 'Cadmium', 'mass': 112.41},
    {'symbol': 'In', 'name': 'Indium', 'mass': 114.82},
    {'symbol': 'Sn', 'name': 'Tin', 'mass': 118.69},
    {'symbol': 'Sb', 'name': 'Antimony', 'mass': 121.75},
    {'symbol': 'I', 'name': 'Iodine', 'mass': 126.9045},
    {'symbol': 'Te', 'name': 'Tellurium', 'mass': 127.60},
    {'symbol': 'Xe', 'name': 'Xenon', 'mass': 131.30},
    {'symbol': 'Cs', 'name': 'Cesium', 'mass': 132.9054},
    {'symbol': 'Ba', 'name': 'Barium', 'mass': 137.33},
    {'symbol': 'La', 'name': 'Lanthanum', 'mass': 138.9055},
    {'symbol': 'Ce', 'name': 'Cerium', 'mass': 140.12},
    {'symbol': 'Pr', 'name': 'Praseodymium', 'mass': 140.9077},
    {'symbol': 'Nd', 'name': 'Neodymium', 'mass': 144.24},
    {'symbol': 'Pm', 'name': 'Promethium', 'mass': (145)},
    {'symbol': 'Sm', 'name': 'Samarium', 'mass': 150.4},
    {'symbol': 'Eu', 'name': 'Europium', 'mass': 151.96},
    {'symbol': 'Gd', 'name': 'Gadolinium', 'mass': 157.25},
    {'symbol': 'Tb', 'name': 'Terbium', 'mass': 158.9254},
    {'symbol': 'Dy', 'name': 'Dysprosium', 'mass': 162.50},
    {'symbol': 'Ho', 'name': 'Holmium', 'mass': 164.9304},
    {'symbol': 'Er', 'name': 'Erbium', 'mass': 167.26},
    {'symbol': 'Tm', 'name': 'Thulium', 'mass': 168.9342},
    {'symbol': 'Yb', 'name': 'Ytterbium', 'mass': 173.04},
    {'symbol': 'Lu', 'name': 'Lutetium', 'mass': 174.967},
    {'symbol': 'Hf', 'name': 'Hafnium', 'mass': 178.49},
    {'symbol': 'Ta', 'name': 'Tantalum', 'mass': 180.9479},
    {'symbol': 'W', 'name': 'Tungsten', 'mass': 183.85},
    {'symbol': 'Re', 'name': 'Rhenium', 'mass': 186.207},
    {'symbol': 'Os', 'name': 'Osmium', 'mass': 190.2},
    {'symbol': 'Ir', 'name': 'Iridium', 'mass': 192.22},
    {'symbol': 'Pt', 'name': 'Platinum', 'mass': 195.09},
    {'symbol': 'Au', 'name': 'Gold', 'mass': 196.9665},
    {'symbol': 'Hg', 'name': 'Mercury', 'mass': 200.59},
    {'symbol': 'Tl', 'name': 'Thallium', 'mass': 204.37},
    {'symbol': 'Pb', 'name': 'Lead', 'mass': 207.2},
    {'symbol': 'Bi', 'name': 'Bismuth', 'mass': 208.9804},
    {'symbol': 'Po', 'name': 'Polonium', 'mass': (209)},
    {'symbol': 'At', 'name': 'Astatine', 'mass': (210)},
    {'symbol': 'Rn', 'name': 'Radon', 'mass': (222)},
    {'symbol': 'Fr', 'name': 'Francium', 'mass': (223)},
    {'symbol': 'Ra', 'name': 'Radium', 'mass': 226.0254},
    {'symbol': 'Ac', 'name': 'Actinium', 'mass': 227.0278},
    {'symbol': 'Pa', 'name': 'Protactinium', 'mass': 231.0359},
    {'symbol': 'Th', 'name': 'Thorium', 'mass': 232.0381},
    {'symbol': 'Np', 'name': 'Neptunium', 'mass': 237.0482},
    {'symbol': 'U', 'name': 'Uranium', 'mass': 238.029},
    {'symbol': 'Pu', 'name': 'Plutonium', 'mass': (242)},
    {'symbol': 'Am', 'name': 'Americium', 'mass': (243)},
    {'symbol': 'Bk', 'name': 'Berkelium', 'mass': (247)},
    {'symbol': 'Cm', 'name': 'Curium', 'mass': (247)},
    {'symbol': 'No', 'name': 'Nobelium', 'mass': (250)},
    {'symbol': 'Cf', 'name': 'Californium', 'mass': (251)},
    {'symbol': 'Es', 'name': 'Einsteinium', 'mass': (252)},
    {'symbol': 'Hs', 'name': 'Hassium', 'mass': (255)},
    {'symbol': 'Mt', 'name': 'Meitnerium', 'mass': (256)},
    {'symbol': 'Fm', 'name': 'Fermium', 'mass': (257)},
    {'symbol': 'Md', 'name': 'Mendelevium', 'mass': (258)},
    {'symbol': 'Lr', 'name': 'Lawrencium', 'mass': (260)},
    {'symbol': 'Rf', 'name': 'Rutherfordium', 'mass': (261)},
    {'symbol': 'Bh', 'name': 'Bohrium', 'mass': (262)},
    {'symbol': 'Db', 'name': 'Dubnium', 'mass': (262)},
    {'symbol': 'Sg', 'name': 'Seaborgium', 'mass': (263)},
]

ELEMENTS = iterables.Table(_elements)

_nucleons = {
    element['symbol']: round(element['mass'])
    for element in ELEMENTS
}

def elements(
    mass: typing.Iterable,
    charge: typing.Iterable,
) -> typing.List[str]:
    """The elemental species symbols, based on masses and charges."""
    _mass = numpy.array(list(iterables.whole(mass)))
    _charge = numpy.array(list(iterables.whole(charge)))
    if len(_mass) != len(_charge):
        message = (
            f"Length of mass ({len(_mass)})"
            f" must equal length of charge ({len(_charge)})"
        )
        raise TypeError(message)
    mass_idx = get_mass_indices(_nucleons, _mass)
    bases = [list(_nucleons.keys())[m] for m in mass_idx]
    signs = [('+' if i > 0 else '-') * abs(int(i)) for i in _charge]
    return [f"{b}{c}" for b, c in zip(bases, signs)]


class MassValueError(Exception):
    """The given mass does not correspond to a known element."""
    def __init__(self, value: int) -> None:
        self.value = value

    def __str__(self) -> str:
        return f"Cannot find an element with atomic mass {self.value}"


def get_mass_indices(
    nucleons: typing.Dict[str, int],
    targets: typing.Iterable,
) -> list:
    """Get the indices in `ELEMENTS` corresponding to the given masses."""
    def get_index(this: list, that: int):
        try:
            return this.index(that)
        except ValueError as err:
            raise MassValueError(that) from err

    values = list(nucleons.values())
    return [get_index(values, round(target)) for target in targets]


def _show_constants():
    """Print all defined physical constants."""
    for key, data in CONSTANTS.items(aliased=True):
        print(f"{key}: {data['info']}")
        for system in ('mks', 'cgs'):
            value = data[system]['value']
            unit = data[system]['unit']
            print(f"\t{system}: {value} [{unit}]")
        print()


def _show_elements():
    """Print all known chemical elements."""
    ELEMENTS.show(names=['symbol', 'name', 'mass'])


def show(*subsets):
    """Print information about physical quantities defined in this module."""
    printers = {
        'constants': _show_constants,
        'elements': _show_elements,
    }
    if not subsets:
        subsets = tuple(printers)
    for subset in subsets:
        if subset in printers:
            printers[subset]()
        else:
            print(f"Nothing to print for {subset!r}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description=show.__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        '--constants',
        help=_show_constants.__doc__,
        action='store_true',
    )
    parser.add_argument(
        '--elements',
        help=_show_elements.__doc__,
        action='store_true',
    )
    args = vars(parser.parse_args())
    subsets = (k for k, v in args.items() if v)
    show(*subsets)





import abc
import collections
import collections.abc
import contextlib
import numbers
import operator as standard
import typing

import numpy

from goats.core import aliased
from goats.core import axis
from goats.core import computed
from goats.core import constant
from goats.core import datafile
from goats.core import interpolation
from goats.core import iterables
from goats.core import metadata
from goats.core import metric
from goats.core import observed
from goats.core import physical
from goats.core import reference
from goats.core import symbolic
from goats.core import variable


T = typing.TypeVar('T')


class Parameters(collections.abc.Sequence, iterables.ReprStrMixin):
    """The parameters of an `~observing.Quantity`."""

    def __init__(self, *names: str) -> None:
        self._names = self._init(*names)

    def _init(self, *args):
        names = iterables.unique(*iterables.unwrap(args, wrap=tuple))
        if not names or all(isinstance(name, str) for name in names):
            return names
        raise TypeError(
            f"Can't initialize instance of {type(self)}"
            f" with {names!r}"
        )

    __abs__ = metadata.identity(abs)
    """Called for abs(self)."""
    __pos__ = metadata.identity(standard.pos)
    """Called for +self."""
    __neg__ = metadata.identity(standard.neg)
    """Called for -self."""

    __add__ = metadata.identity(standard.add)
    """Called for self + other."""
    __sub__ = metadata.identity(standard.sub)
    """Called for self - other."""

    def merge(a, *others):
        """Return the unique axis names in order."""
        names = list(a._names)
        for b in others:
            if isinstance(b, Parameters):
                names.extend(b._names)
        return Parameters(*set(names))

    __mul__ = merge
    """Called for self * other."""
    __rmul__ = merge
    """Called for other * self."""
    __truediv__ = merge
    """Called for self / other."""

    def __pow__(self, other):
        """Called for self ** other."""
        if isinstance(other, numbers.Real):
            return self
        return NotImplemented

    def __eq__(self, other):
        """True if self and other represent the same axes."""
        return (
            isinstance(other, Parameters) and other._names == self._names
            or (
                isinstance(other, str)
                and len(self) == 1
                and other == self._names[0]
            )
            or (
                isinstance(other, typing.Iterable)
                and len(other) == len(self)
                and all(i in self for i in other)
            )
        )

    def __hash__(self):
        """Support use as a mapping key."""
        return hash(self._names)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._names)

    def __getitem__(self, __i: typing.SupportsIndex):
        """Called for index-based access."""
        return self._names[__i]

    def __str__(self) -> str:
        return f"[{', '.join(repr(name) for name in self._names)}]"


class Quantity(variable.Quantity):
    """A quantity with one or more name(s), a unit, and axes."""

    _parameters: Parameters=None

    def __init__(self, __data, **meta) -> None:
        super().__init__(__data, **meta)
        parsed = self.parse_attrs(__data, meta, parameters=())
        self._parameters = Parameters(parsed['parameters'])
        self.meta.register('parameters')
        self.display.register('parameters')
        self.display['__str__'].append("parameters={parameters}")
        self.display['__repr__'].append("parameters={parameters}")

    def parse_attrs(self, this, meta: dict, **targets):
        if (
            isinstance(this, variable.Quantity)
            and not isinstance(this, Quantity)
        ): # barf
            meta.update({k: getattr(this, k) for k in ('unit', 'name', 'axes')})
            this = this.data
        return super().parse_attrs(this, meta, **targets)

    @property
    def parameters(self):
        """The optional parameters that define this observing quantity."""
        return Parameters(self._parameters)


class Dataset(abc.ABC, typing.Generic[T]):
    """Abstract base class for observing-related data objects."""

    def __init__(self, __type: typing.Type[T], source) -> None:
        self._type = __type
        self._data = self._type(source)

    @abc.abstractmethod
    def get_axes(self, system: str=None) -> axis.Interface:
        """Get the available axis-managing objects."""
        return axis.Interface(self.data, system=system)

    @abc.abstractmethod
    def get_variables(self, system: str=None) -> variable.Interface:
        """Get the available variable quantities."""
        return variable.Interface(self.data, system=system)

    @abc.abstractmethod
    def get_constants(self) -> constant.Interface:
        """Get the available constant quantities."""
        return constant.Interface()

    @property
    def data(self):
        """The object that manages access to this dataset's data."""
        return self._data

    def readfrom(self, source):
        """Update the data source."""
        self._data = self._type(source)
        return self


class Quantities(collections.abc.Collection):
    """An interface to quantities required to make observations."""

    def __init__(self, dataset: Dataset, system: str=None) -> None:
        self._dataset = dataset
        self._system = metric.System(system or 'mks')
        self._axes = None
        self._variables = None
        self._constants = None
        self._names = None
        self._primary = None
        self._derived = None
        self._functions = None
        self._assumptions = None

    def __contains__(self, __x) -> bool:
        return __x in self.names or any(__x in name for name in self.names)

    def __iter__(self) -> typing.Iterator:
        return iter(self.names)

    def __len__(self) -> int:
        return len(self.names)

    @property
    def axes(self):
        """The available axis-managing objects."""
        if self._axes is None:
            self._axes = self._dataset.get_axes(self.system)
        return self._axes

    @property
    def variables(self):
        """The available variable quantities."""
        if self._variables is None:
            self._variables = self._dataset.get_variables(self.system)
        return self._variables

    @property
    def assumptions(self):
        """The default physical assumptions."""
        if self._assumptions is None:
            assumptions = {
                k: v
                for k, v in self.constants.items(aliased=True)
                if isinstance(v, constant.Assumption)
            } if self.constants else {}
            self._assumptions = aliased.Mapping(assumptions)
        return self._assumptions

    @property
    def constants(self):
        """The available constant quantities."""
        if self._constants is None:
            self._constants = self._dataset.get_constants()
        return self._constants

    @property
    def system(self):
        """The metric system to use for observations."""
        return self._system

    @property
    def names(self):
        """The names of all available quantities."""
        if self._names is None:
            self._names = self.primary + self.derived
        return self._names

    @property
    def primary(self):
        """The names of all available primary quantities.
        
        A primary quantity is a quantity that comes directly from the dataset.
        """
        if self._primary is None:
            self._primary = tuple(self.variables.keys(aliased=True))
        return self._primary

    @property
    def derived(self):
        """The names of all available derived quantities.
        
        A derived quantity is a quantity that is the result of a defined
        function of one or more primary or derived quantities, and zero or more
        physical assumptions.
        """
        if self._derived is None:
            self._derived = tuple(self.functions.keys(aliased=True))
        return self._derived

    @property
    def functions(self):
        """The names of functions for computing derived quantities."""
        if self._functions is None:
            self._functions = computed.Interface(self.axes, self.variables)
        return self._functions

    def get_observable(self, key: str):
        """Get the observable quantity corresponding to `key`."""
        for this in (self.variables, self.functions):
            if key in this:
                return this[key]

    # TODO: Refactor `get_unit` and `get_axes` to reduce overlap.

    def get_unit(self, key: str):
        """Determine the unit corresponding to `key`."""
        if key in self.variables:
            return self.variables[key].unit
        if key in self.functions:
            return self.functions[key].unit
        s = str(key)
        expression = symbolic.Expression(reference.NAMES.get(s, s))
        term = expression[0]
        this = self.get_unit(term.base) ** term.exponent
        if len(expression) > 1:
            for term in expression[1:]:
                this *= self.get_unit(term.base) ** term.exponent
        return metadata.Unit(this)

    def get_axes(self, key: str):
        """Determine the axes corresponding to `key`."""
        if key in self.variables:
            return self.variables[key].axes
        if key in self.functions:
            return self.functions[key].axes
        s = str(key)
        expression = symbolic.Expression(reference.NAMES.get(s, s))
        term = expression[0]
        this = self.get_axes(term.base)
        if len(expression) > 1:
            for term in expression[1:]:
                this *= self.get_axes(term.base)
        return metadata.Axes(this)

    def compute_index(self, key: str, **constraints) -> axis.Quantity:
        """Compute the axis-indexing object for `key`."""
        if key not in self.axes:
            raise ValueError(f"No axis corresponding to {key!r}") from None
        if key not in constraints:
            return self.axes[key].index()
        this = constraints[key]
        if isinstance(this, axis.Quantity):
            return this
        return self.axes[key].index(*iterables.whole(this))

    def compute_value(self, key: str, **constraints) -> physical.Scalar:
        """Create a parameter value for `key`."""
        if key not in self.constants and key not in constraints:
            raise ValueError(
                f"No parameter corresponding to {key!r}"
            ) from None
        if key in constraints:
            return self._compute_value(constraints[key])
        return self._compute_value(self.constants[key])

    def _compute_value(self, this):
        """Compute a parameter value."""
        # TODO: Generalize beyond scalar parameters.
        scalar = physical.scalar(this)
        unit = self.system.get_unit(unit=scalar.unit)
        return scalar[unit]


class Application:
    """A general observing application."""

    def __init__(self, interface: Quantities) -> None:
        self.interface = interface
        """The interface to an observer's dataset."""
        self.user = None
        """The current set of user constraints."""
        self._coordinates = None
        self._cache = {}

    def __contains__(self, __k: str):
        """True if the named constraint affects this application."""
        # Should this check all dimensions and parameters or only user
        # constraints?
        return __k in self.user

    def apply(self, **constraints):
        """Update the observing constraints.
        
        This method will apply the given constraints when computing axis indices
        and parameter values within the context of this application. It will
        also clear the corresponding items from the instance cache.
        """
        if self.user is None:
            self.user = {}
        for key in constraints:
            for subset in self._cache.values():
                if key in subset:
                    del subset[key]
        self.user.update(constraints)
        return self

    def reset(self, *keys: str):
        """Reset constraints to their default values.
        
        Parameters
        ----------
        *keys : string
            Zero or more names of constraints to reset. This method will also
            remove all named items from the instance cache By default, this
            method will reset all constraints and clear the cache.
        """
        if not keys:
            self.user = {}
            self._cache = {}
        else:
            subsets = self._cache.values()
            for key in keys:
                if key in self.user:
                    del self.user[key]
                for subset in subsets:
                    if key in subset:
                        del subset[key]
        return self

    def observe(self, key: str) -> Quantity:
        """Create an observation within the context of this application."""
        result = self._observe(key)
        if any(alias in self.coordinates for alias in result.name):
            # This is an axis-reference quantity.
            return self._subscript(result)
        needed = self._compute_interpolants(result)
        if not needed:
            # There are no axes over which to interpolate.
            return self._subscript(result)
        return self.interpolate(result, needed)

    def _observe(self, key: str) -> Quantity:
        """Internal observing logic."""
        expression = symbolic.Expression(reference.NAMES.get(key, key))
        term = expression[0]
        result = self.get_observable(term.base)
        if len(expression) == 1:
            # We don't need to multiply or divide quantities.
            if term.exponent == 1:
                # We don't even need to raise this quantity to a power.
                return result
            return result ** term.exponent
        q0 = result ** term.exponent
        if len(expression) > 1:
            for term in expression[1:]:
                result = self.get_observable(term.base)
                q0 *= result ** term.exponent
        return q0

    def evaluate(self, q) -> Quantity:
        """Create an observing quantity based on the given quantity."""
        if isinstance(q, computed.Quantity):
            parameters = [
                parameter for parameter in q.parameters
                if parameter in self.interface.constants
            ]
            return Quantity(self.compute(q), parameters=parameters)
        if isinstance(q, variable.Quantity):
            return Quantity(self.process(q))
        raise ValueError(f"Unknown quantity: {q!r}") from None

    def process(self, q: variable.Quantity) -> variable.Quantity:
        """Compute observer-specific updates to a variable quantity.
        
        The default implementation immediately returns `q`.
        """
        return q

    def compute(self, q: computed.Quantity) -> variable.Quantity:
        """Determine dependencies and compute the result of this function."""
        dependencies = {p: self.get_dependency(p) for p in q.parameters}
        return q(**dependencies)

    def get_dependency(self, key: str):
        """Get the named constant or variable quantity."""
        if this := self.get_observable(key):
            return this
        return self.get_value(key)

    def get_observable(self, key: str):
        """Retrieve and evaluate an observable quantity."""
        if quantity := self.interface.get_observable(key):
            return self.evaluate(quantity)

    def _subscript(self, q: variable.Quantity, *axes: str):
        """Extract a subset of this quantity."""
        if not axes:
            return q[tuple(self.get_index(a, slice(None)) for a in q.axes)]
        indices = [
            self.get_index(a, slice(None))
            if a in axes else slice(None)
            for a in q.axes
        ]
        return q[tuple(indices)]

    def _compute_interpolants(self, q: variable.Quantity):
        """Determine the coordinate axes over which to interpolate."""
        coordinates = {}
        for a in q.axes:
            idx = self.get_index(a)
            if idx and idx.unit is not None:
                contained = [
                    self.coordinates[a].array_contains(target)
                    for target in idx.values
                ]
                if not numpy.all(contained):
                    coordinates[a] = {
                        'targets': numpy.array(idx.values),
                        'reference': self.coordinates[a],
                    }
        return coordinates

    def interpolate(
        self,
        q: variable.Quantity,
        coordinates: typing.Dict[str, typing.Dict[str, typing.Any]],
    ) -> variable.Quantity:
        """Internal interpolation logic."""
        array = None
        for coordinate in coordinates.values():
            array = self._interpolate_coordinate(
                q,
                coordinate['targets'],
                coordinate['reference'],
                axis=coordinate.get('axis'),
                workspace=array,
            )
        meta = {k: getattr(q, k, None) for k in q.meta.parameters}
        return type(q)(array, **meta)

    def _interpolate_coordinate(
        self,
        q: variable.Quantity,
        targets: numpy.ndarray,
        reference: variable.Quantity,
        axis: int=None,
        workspace: numpy.ndarray=None,
    ) -> numpy.ndarray:
        """Interpolate a variable array based on a known coordinate."""
        array = numpy.array(q) if workspace is None else workspace
        indices = (q.axes.index(d) for d in reference.axes)
        dst, src = zip(*enumerate(indices))
        reordered = numpy.moveaxis(array, src, dst)
        interpolated = interpolation.apply(
            reordered,
            numpy.array(reference),
            targets,
            axis=axis,
        )
        return numpy.moveaxis(interpolated, dst, src)

    @property
    def coordinates(self):
        """The reference quantities for dataset coordinate axes."""
        if self._coordinates is None:
            coordinates = self._build_coordinates()
            self._coordinates = aliased.Mapping(coordinates)
        return self._coordinates

    def _build_coordinates(self):
        """Helper for `~coordinates` property. Extracted for overloading."""
        return {
            k: self.interface.variables.get(k)
            for k in self.interface.axes.keys(aliased=True)
        }

    def get_index(self, key: str, default: T=None) -> axis.Quantity:
        """Get the axis-indexing object for `key`."""
        if 'indices' not in self._cache:
            self._cache['indices'] = {}
        if key in self._cache['indices']:
            return self._cache['indices'][key]
        with contextlib.suppress(ValueError):
            idx = self.interface.compute_index(key, **self.user)
            self._cache['indices'][key] = idx
            return idx
        return default

    def get_value(self, key: str, default: T=None) -> physical.Scalar:
        """Get the parameter value correpsonding to `key`."""
        if 'values' not in self._cache:
            self._cache['values'] = {}
        if key in self._cache['values']:
            return self._cache['values'][key]
        with contextlib.suppress(ValueError):
            val = self.interface.compute_value(key, **self.user)
            self._cache['values'][key] = val
            return val
        return default


class Implementation:
    """The base observing implementation."""

    def __init__(
        self,
        name: str,
        interface: Quantities,
    ) -> None:
        """
        Initialize this instance.

        Parameters
        ----------
        name : string
            The name of the quantity to observe.

        interface : `~Interface`
            The interface to all observing-related quantities.
        """
        self.name = name
        self.interface = interface
        self._cache = {}

    # TODO:
    # - refactor `get_unit` and `get_axes` to reduce overlap.
    # - redefine `_cache` as an aliased mapping.

    def get_unit(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Return the unit corresponding to `name`."""
        if 'unit' not in self._cache:
            self._cache['unit'] = {}
        if name in self._cache['unit']:
            return self._cache['unit'][name]
        unit = self._lookup_unit(name)
        self._cache['unit'][name] = unit
        return unit

    def _lookup_unit(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Determine the unit corresponding to `name`."""
        if isinstance(name, str):
            return self.interface.get_unit(name)
        unit = (self.interface.get_unit(key) for key in name)
        return next(unit, None)

    def get_axes(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Axes:
        """Return the axes corresponding to `name`."""
        if 'axes' not in self._cache:
            self._cache['axes'] = {}
        if name in self._cache['axes']:
            return self._cache['axes'][name]
        axes = self._lookup_axes(name)
        self._cache['axes'][name] = axes
        return axes

    def _lookup_axes(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Determine the axes corresponding to `name`."""
        if isinstance(name, str):
            return self.interface.get_axes(name)
        axes = (self.interface.get_axes(key) for key in name)
        return next(axes, None)

    def apply(
        self,
        application: Application,
        unit: metadata.UnitLike=None,
    ) -> observed.Quantity:
        """Apply an observing context to the target quantity."""
        result = application.observe(self.name)
        indices = {k: application.get_index(k) for k in result.axes}
        scalars = {k: application.get_value(k) for k in result.parameters}
        context = observed.Context(indices, scalars=scalars)
        if unit:
            return observed.Quantity(result[unit], context)
        return observed.Quantity(result, context)



import contextlib
import pathlib
import typing

import numpy
import numpy.typing

from ..core import (
    axis,
    constant,
    datafile,
    fundamental,
    iotools,
    index,
    iterables,
    measurable,
    metadata,
    metric,
    numerical,
    observer,
    observing,
    physical,
    reference,
    variable,
)
from .runtime import BaseTypesH


basetypes = BaseTypesH()


T = typing.TypeVar('T')


class Axes(axis.Interface):
    """Interface to EPREM axis-indexing objects."""

    def __init__(
        self,
        data: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        super().__init__(data, system)
        self._time = None
        self._shell = None
        self._species = None
        self._energy = None
        self._mu = None
        defined = {
            'time': self.time,
            'shell': self.shell,
            'species': self.species,
            'energy': self.energy,
            'mu': self.mu,
        }
        self.update(defined)

    @property
    def shell(self):
        """Indexer for the EPREM shell dimension."""
        if self._shell is None:
            def method(targets, unit):
                # NOTE: The presence of `unit` is a hack because 'shell'
                # currently gets a unit of '1' even though it should probably be
                # None. This hack is due to the design of `axis.Quantity`.
                return index.Data(targets)
            self._shell = axis.Indexer(method, len(self.variables['shell']))
        return self._shell

    @property
    def time(self):
        """Indexer for the EPREM time dimension."""
        if self._time is None:
            this = self.variables['time']
            self._time = axis.Indexer(self._build_coordinate(this), len(this))
        return self._time

    @property
    def mu(self):
        """Indexer for the EPREM pitch-angle dimension."""
        if self._mu is None:
            this = self.variables['mu']
            self._mu = axis.Indexer(self._build_coordinate(this), len(this))
        return self._mu

    @property
    def energy(self):
        """Indexer for the EPREM energy dimension."""
        if self._energy is None:
            this = self.variables['energy']
            def method(
                targets,
                unit: metadata.UnitLike,
                species: typing.Union[str, int]=0,
            ) -> index.Data:
                s = self.species.compute([species]).points
                t = (
                    numpy.squeeze(targets[s, :])
                    if getattr(targets, 'ndim', None) == 2
                    else targets
                )
                compute = self._build_coordinate(numpy.squeeze(this[s, :]))
                return compute(t, unit)
            self._energy = axis.Indexer(method, this.shape[1])
        return self._energy

    @property
    def species(self):
        """Indexer for the EPREM species dimension."""
        if self._species is None:
            mass = self.variables['mass']['nuc']
            charge = self.variables['charge']['e']
            symbols = fundamental.elements(mass, charge)
            def method(targets):
                indices = [
                    symbols.index(target)
                    if isinstance(target, str) else int(target)
                    for target in targets
                ]
                return index.Data(indices, values=symbols)
            self._species = axis.Indexer(method, len(symbols))
        return self._species

    def _build_coordinate(self, this: variable.Quantity):
        """Create coordinate-like axis data from the given variable."""
        def method(targets, unit: typing.Union[str, metadata.Unit]):
            # Convert the reference variable quantity to the default unit.
            converted = this[unit]
            if not targets:
                # If there are no target values, we assume the user wants the
                # entire axis.
                return index.Data(
                    range(len(converted)),
                    values=numpy.array(converted),
                )
            if all(isinstance(t, typing.SupportsIndex) for t in targets):
                # All the target values are already indices.
                return index.Data(targets, values=numpy.array(converted))
            measured = measurable.measure(targets)
            if measured.unit != '1':
                # If the targets include a dimensioned unit, we want to
                # initialize the array with that unit.
                array = physical.Array(measured.values, unit=measured.unit)
            else:
                # If the measured unit is dimensionless, it could be because the
                # targets truly are dimensionless or because the user wants to
                # use the default unit. Since we have no choice but to assume
                # that the calling object (probably an instance of
                # `core.axis.Quantity`) passed an appropriate default unit,
                # which may be dimensionless, the default unit is the
                # appropriate unit for both cases.
                array = physical.Array(measured.values, unit=unit)
            if array.unit | converted.unit: # Could also use try/except
                array = array[converted.unit]
            values = numpy.array(array)
            indices = [
                numerical.find_nearest(converted, float(value)).index
                for value in values
            ]
            return index.Data(indices, values=values)
        return method


Instance = typing.TypeVar('Instance', bound='Dataset')


class Dataset(observing.Dataset):
    """Interface to an EPREM dataset."""

    def __init__(
        self,
        datapath: iotools.ReadOnlyPath,
        confpath: iotools.ReadOnlyPath,
    ) -> None:
        super().__init__(datafile.Interface, datapath)
        self._datapath = datapath
        self._confpath = confpath

    def get_axes(self, system: str=None) -> axis.Interface:
        return Axes(self.data, system=system)

    def get_variables(self, system: str=None) -> variable.Interface:
        return variable.Interface(self.data, system=system)

    def get_constants(self) -> constant.Interface:
        return runtime.Arguments(config_path=self.confpath)

    def readfrom(
        self: Instance,
        datapath: iotools.ReadOnlyPath,
        confpath: iotools.ReadOnlyPath=None,
    ) -> Instance:
        self._datapath = datapath
        if confpath:
            self._confpath = confpath
        return super().readfrom(datapath)

    @property
    def confpath(self) -> iotools.ReadOnlyPath:
        """The full path to this dataset's runtime parameter file."""
        return self._confpath

    @property
    def datapath(self) -> iotools.ReadOnlyPath:
        """The path to this dataset."""
        return self._datapath


class Application(observing.Application):
    """The EPREM-specific observing context."""

    def interpolate(
        self,
        q: variable.Quantity,
        coordinates: typing.Dict[str, typing.Dict[str, typing.Any]],
    ) -> variable.Quantity:
        base = super().interpolate(q, coordinates)
        # We only want to subscript the uninterpolated axes.
        interpolated = [
            'shell' if d == 'radius' else d
            for d in coordinates
        ]
        axes = list(set(q.axes) - set(interpolated))
        return self._subscript(base, *axes)

    def _build_coordinates(self):
        base = super()._build_coordinates()
        grid = {
            (k, *self.interface.variables.alias(k, include=True)):
            self.interface.variables[k]
            for k in {'radius', 'theta', 'phi'}
        }
        return {**base, **grid}

    _axes = {
        'time': 0,
        'energy': 1,
        'mu': 1,
    }

    def _compute_interpolants(self, q: variable.Quantity):
        base = {
            k: {**c, 'axis': self._axes.get(k)}
            for k, c in super()._compute_interpolants(q).items()
        }
        if 'shell' not in q.axes:
            # The rest of this method deals with radial interpolation, which
            # only applies when 'shell' is one of the target quantity's axes.
            return base
        for key in reference.ALIASES['radius']:
            if values := self.get_value(key):
                try:
                    iter(values)
                except TypeError:
                    floats = [float(values)]
                else:
                    floats = [float(value) for value in values]
                base['radius'] = {
                    'targets': numpy.array(floats),
                    'reference': self.coordinates['radius'],
                    'axis': 1,
                }
        return base


Instance = typing.TypeVar('Instance', bound='Observer')


class Observer(observer.Interface, iterables.ReprStrMixin):
    """Base class for EPREM observers."""

    _templates: typing.Iterable[typing.Callable] = None

    _unobservable = [
        'preEruption',
        'phiOffset',
    ]

    def __init__(
        self,
        __id: int,
        config: iotools.PathLike,
        source: iotools.PathLike=pathlib.Path.cwd(),
        system: str='mks',
    ) -> None:
        self._id = __id
        datapath = self._build_datapath(source)
        confpath = self._build_confpath(config, directory=datapath.parent)
        super().__init__(
            Dataset(datapath, confpath),
            *self._unobservable,
            system=system,
            apply=Application,
        )
        self._axes = None

    def reset(
        self: Instance,
        source: iotools.PathLike=None,
        config: iotools.PathLike=None
    ) -> Instance:
        datapath = (
            self._build_datapath(source) if source is not None
            else self.datapath
        )
        if config is None:
            return super().reset(source=datapath)
        confpath = self._build_confpath(config, directory=datapath.parent)
        self._confpath = confpath
        self._axes = None
        return super().reset(source=datapath, confpath=confpath)

    def _build_datapath(self, directory: iotools.PathLike):
        """Create the path to the dataset from `directory`."""
        this = iotools.ReadOnlyPath(directory or pathlib.Path.cwd())
        if this.is_dir():
            path = iotools.find_file_by_template(
                self._templates,
                self._id,
                directory=this,
            )
            with contextlib.suppress(TypeError):
                # Couldn't initialize with `path` for some reason.
                return iotools.ReadOnlyPath(path)
        raise TypeError(
            f"Can't create path to dataset from {directory!r}"
        ) from None

    def _build_confpath(
        self,
        config: iotools.PathLike,
        directory: iotools.PathLike=None,
    ) -> iotools.ReadOnlyPath:
        """Create the path to the configuration file."""
        # Create a `Path` but don't expand or resolve it.
        path = pathlib.Path(config)
        if directory and path.name == config:
            # There is a directory to check and `config` is just a filename.
            full = iotools.ReadOnlyPath(directory) / config
            if full.exists():
                return full
        # Expand and resolve path.
        this = iotools.ReadOnlyPath(config)
        if this.exists():
            # The absolute path exists, so return it.
            return this
        raise ValueError(
            f"Can't create path to configuration file from {config!r}"
        ) from None

    @property
    def confpath(self) -> iotools.ReadOnlyPath:
        """The full path to this observer's runtime parameter file."""
        return self._data.confpath

    @property
    def datapath(self) -> iotools.ReadOnlyPath:
        """The path to this observer's dataset."""
        return self._data.datapath

    @property
    def radius(self):
        """The time-dependent radius values in this observer's dataset."""
        return self.quantities.variables['radius']

    @property
    def theta(self):
        """The time-dependent theta values in this observer's dataset."""
        return self.quantities.variables['theta']

    @property
    def phi(self):
        """The time-dependent phi values in this observer's dataset."""
        return self.quantities.variables['phi']

    @property
    def time(self):
        """The time values in this observer's dataset."""
        return self.quantities.axes['time'].reference

    @property
    def shell(self):
        """The shell values in this observer's dataset."""
        return self.quantities.axes['shell'].reference

    @property
    def species(self):
        """The species values in this observer's dataset."""
        return self.quantities.axes['species'].reference

    @property
    def energy(self):
        """The energy values in this observer's dataset."""
        return self.quantities.axes['energy'].reference

    @property
    def mu(self):
        """The pitch-angle cosine values in this observer's dataset."""
        return self.quantities.axes['mu'].reference

    def __str__(self) -> str:
        return str(self.datapath)


class Stream(Observer):
    """An EPREM stream observer."""

    _templates = [
        lambda n: f'obs{n:06}.nc',
        lambda n: f'flux{n:06}.nc',
    ]


class Point(Observer):
    """An EPREM point observer."""

    _templates = [
        lambda n: f'p_obs{n:06}.nc',
    ]




import abc
import collections
import collections.abc
import fractions
import functools
import itertools
import numbers
import re
from operator import attrgetter
import typing

from goats.core import algebraic
from goats.core import iterables


# NOTE: This is an experimental attempt to refactor redundant code in `Term` and
# `OperandFactory`. It is currently not in use.
class Patterns(collections.UserDict):
    """Compiled regular expressions for symbolic operands."""

    rational = r""" # Modeled after `fractions._RATIONAL_FORMAT`
        [-+]?                 # an optional sign, ...
        (?=\d|\.\d)           # ... only if followed by <digit> or .<digit>
        \d*                   # and a possibly empty numerator
        (?:                   # followed by ...
            (?:/\d+?)         # ... an optional denominator
        |                     # OR
            (?:\.\d*)?        # ... an optional fractional part,
            (?:[eE][-+]?\d+)? #     and an optional exponent
        )
    """
    base = r"""
        [a-zA-Z#_]+ # one or more accepted non-digit character(s)
        \d*         # followed by optional digits
    """

    def __init__(
        self,
        multiplication: str='*',
        division: str='/',
        opening: str='(',
        closing: str=')',
        raising: str='^',
    ) -> None:
        multiply = fr'\{multiplication}'
        divide = fr'\{division}'
        exponent = fr'\{raising}{self.rational}'
        patterns = {
            'multiply': (
                fr'(?<!{divide})(\s*{multiply}\s*)(?!{divide})'
            ),
            'divide': (
                fr'(?<!{multiply})(\s*{divide}\s*)(?!{multiply})'
            ),
            'sqrt': (r'\s*sqrt\s*'),
            'constant': (
                fr'(?P<coefficient>{self.rational})'
                fr'(?P<exponent>{exponent})?'
            ),
            'variable': (
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>{self.base})'
                fr'(?P<exponent>{exponent})?'
            ),
            'expression': (
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>\{opening}.+?\{closing})'
                fr'(?P<exponent>{exponent})?'
            ),
            'exponent': exponent,
            'opening': fr'\{opening}',
            'closing': fr'\{closing}',
            'raising': fr'\{raising}',
        }
        self._compiled = {}
        super().__init__(patterns)

    def __getitem__(self, __k: str):
        if __k in self._compiled:
            return self._compiled[__k]
        if __k in self.data:
            compiled = re.compile(self.data[__k], re.VERBOSE)
            self._compiled[__k] = compiled
            return compiled
        raise KeyError(f"No pattern for {__k}") from None

    def __setitem__(self, __k: str, __v: str) -> None:
        self._compiled.pop(__k, None)
        return super().__setitem__(__k, __v)

    def __delitem__(self, __k: str) -> None:
        self._compiled.pop(__k, None)
        return super().__delitem__(__k)


class Part(abc.ABC, iterables.ReprStrMixin):
    """Base class for parts of a symbolic expression."""

    __slots__ = ()


class Operator(Part):
    """An operator in a symbolic expression."""

    def __init__(self, operation: str) -> None:
        self.operation = operation

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.operation

    def __hash__(self) -> str:
        """Compute instance hash (e.g., for use as `dict` key)."""
        return hash(self.operation)

    def __eq__(self, other) -> bool:
        """True if two operators represent the same operation."""
        if isinstance(other, Operator):
            return other.operation == self.operation
        if isinstance(other, str):
            return other == self.operation
        return NotImplemented


class Operand(Part):
    """An operand in a symbolic expression.

    Algebraic operands mainly exist to support the `~algebra.Expression` class.
    They may be simple or general, as described below.

    A simple symbolic operand has the form [c]b[^e] or c[^e], where `c` is a
    numerical coefficient, `b` is a string base, and `e` is a numerical
    exponent. Braces ('[]') denote an optional component that defaults to a
    value of 1. The `~algebra.Term` class formally represents simple symbolic
    operands; the form [c]b[^e] corresponds to a variable term and the form
    c[^e] corresponds to a constant term.

    Examples include:
    
    * `'1'`: unity / multiplicative identity
    * `'1^n'` (`n` real): constant equivalent to unity
    * `'m^n'` (`m`, `n` real): arbitrary constant
    * `'V'`: variable 'V' with coefficient 1 and exponent 1
    * `'V^n'` (`n` real): variable 'V' with coefficient 1 and exponent n
    * `'mV'` (`m` real): variable 'V' with coefficient m and exponent 1
    * `'mV^n'` (`m`, `n` real): variable 'V' with coefficient m and exponent n

    Note that the base of a variable term may comprise multiple characters as
    long as it does not begin with a digit, which this class will interpret as
    part of the coefficient.

    A general symbolic operand consists of simpler operands (though not
    necessarily formally simple operands) combined with symbolic operators and
    separators. All formally simple operands are general operands. The following
    are examples of (non-simple) general symbolic operands:

    * `'a * b^2'` <=> `'(a * b^2)'` <=> `'(a * b^2)^1'`
    * `'(a * b^2)^3'`
    * `'(a * b^2)^3/2'`
    * `'((a / b^2)^3 * c)^2'`
    * `'(a / b^2)^3 * c^2'`
    * `'a / (2 * 4b)'`
    * `'(2a * b)^3 / (4 * c)'`

    There are many more ways to construct a general operand than a simple
    operand. This is by design, to support building instances of
    `~symbolic.Expression` with `~symbolic.Parser`.
    """

    def __init__(
        self,
        coefficient: numbers.Real=None,
        base: str=None,
        exponent: numbers.Real=None,
    ) -> None:
        self.coefficient = fractions.Fraction(coefficient or 1)
        """The numerical coefficient."""
        self.base = base or '1'
        """The base term or complex."""
        self.exponent = fractions.Fraction(exponent or 1)
        """The numerical exponent."""

    @property
    def attrs(self):
        """The current coefficient, base, and exponent."""
        return (self.coefficient, self.base, self.exponent)

    def __pow__(self, power):
        """Create a new operand, raised to `power`."""
        coefficient = self.coefficient ** power
        exponent = self.exponent * power
        return type(self)(coefficient, self.base, exponent)

    def __mul__(self, other):
        """Create a new operand, multiplied by `other`."""
        coefficient = self.coefficient * other
        return type(self)(coefficient, self.base, self.exponent)

    __rmul__ = __mul__

    def __eq__(self, other) -> bool:
        """True if two operands' attributes are equal."""
        if not isinstance(other, Operand):
            return NotImplemented
        if not other.base == self.base:
            return False
        return all(
            float(getattr(other, attr)) == float(getattr(self, attr))
            for attr in ('exponent', 'coefficient')
        )

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.format()

    def format(self):
        """Format this operand for printing."""
        if self.coefficient == 1 and self.exponent == 1:
            return self.base
        string = f"({self.base})"
        if self.exponent != 1:
            string = f"{string}^{self.exponent}"
        if self.coefficient != 1:
            string = f"{self.coefficient}{string}"
        return string


class Term(Operand):
    """A symbolic operand with an irreducible base.
    
    In the single-argument form, the user provides only the base quantity by
    positional-only argument. In the triple-argument form, the user provides the
    coefficient, base, and exponent by positional or keyword argument(s), with
    the caveat that if one argument is positional, they must all be positional.
    This restriction prevents ambiguity among the possible double-argument
    combinations.
    """

    # NOTE: Currently redundant with `OperandFactory`.

    rational = r""" # Modeled after `fractions._RATIONAL_FORMAT`
        [-+]?                 # an optional sign, ...
        (?=\d|\.\d)           # ... only if followed by <digit> or .<digit>
        \d*                   # and a possibly empty numerator
        (?:                   # followed by ...
            (?:/\d+?)         # ... an optional denominator
        |                     # OR
            (?:\.\d*)?        # ... an optional fractional part,
            (?:[eE][-+]?\d+)? #     and an optional exponent
        )
    """
    base = r"""
        [a-zA-Z#_]+ # one or more accepted non-digit character(s)
        \d*         # followed by optional digits
    """
    _base_re = re.compile(fr'({rational}|{base})', re.VERBOSE)

    @typing.overload
    def __init__(self, base: str, /) -> None:
        """Create a symbolic term from a base quantity.
        
        Parameters
        ----------
        base : string
            The base quantity of this symbolic term.
        """

    @typing.overload
    def __init__(
        self,
        coefficient: numbers.Real=1,
        base: str='1',
        exponent: numbers.Real=1,
    ) -> None:
        """Create a symbolic term from coefficient, base, and exponent.
        
        Parameters
        ----------
        coefficient : real number, default=1
            The numerical coefficient to associate with the base quantity.

        base : string, default='1'
            The base quantity of this symbolic term.

        exponent : real number, default=1
            The exponent of the base quantity.
        """

    def __init__(self, *args, **kwargs) -> None:
        if not kwargs:
            # positional argument(s) only
            if not args:
                # use all default values
                return super().__init__()
            if len(args) == 1:
                # single-argument form (positional base only)
                base = args[0]
                self._validate_base(base)
                return super().__init__(base=base)
            if len(args) == 3:
                # triple-argument form
                base = args[1]
                self._validate_base(base)
                return super().__init__(*args)
        if not args:
            # keyword arguments only
            if 'base' in kwargs:
                # NOTE: We don't use `base = kwargs.get('base')` because we
                # don't want to mistakenly treat `None` as the base when
                # comparing to the base RE pattern. We'll let `Operand`
                # determine the default value.
                base = kwargs['base']
                self._validate_base(base)
            return super().__init__(**kwargs)
        raise TypeError(
            f"Can't instantiate {self.__class__} with {args} and {kwargs}."
        ) from None

    def _validate_base(self, base):
        """Raise an exception if `base` does not match the definition."""
        if not self._base_re.fullmatch(str(base)):
            raise ValueError(
                f"Can't create a symbolic term with base {base!r}"
            ) from None

    def __call__(self, value: numbers.Real):
        """Evaluate a variable term at this value.
        
        This method will attempt to substitute `value` for this term's `base`
        attribute. If successful, it will return a constant term that the caller
        may cast to `int` or `float` type.

        Parameters
        ----------
        value : real
            The numerical value at which to evaluate this term.

        Returns
        -------
        `~algebra.Term`
            A new instance of this class equivalent to the constant numerical
            value of this term when `base == value`.
        """
        if not isinstance(value, numbers.Real):
            errmsg = f"Can't evaluate term with value {value!r}"
            raise TypeError(errmsg)
        if self.base != '1':
            coefficient = self.coefficient * (value ** self.exponent)
            return type(self)(coefficient=coefficient)
        errmsg = f"Can't evaluate term with base {self.base!r}"
        raise TypeError(errmsg)

    def format(self, style: str=None):
        """Format this term."""
        coefficient = self._format_coefficient()
        if self.base == '1':
            return f"{coefficient}"
        exponent = self._format_exponent(style)
        return f"{coefficient}{self.base}{exponent}"

    def _format_coefficient(self):
        """Format the coefficient for printing."""
        if self.base != '1' and self.coefficient == 1:
            return ''
        if self._ambiguous_coefficient():
            return f"({self.coefficient})"
        if float(self.coefficient) == int(self.coefficient):
            return str(int(self.coefficient))
        return str(self.coefficient)

    def _ambiguous_coefficient(self):
        """True if this term's coefficient needs to be grouped."""
        return (
            isinstance(self.coefficient, fractions.Fraction)
            and self.base != '1'
            and self.coefficient.denominator != 1
        )

    def _format_exponent(self, style: str):
        """Format the current exponent for printing."""
        if self.base == '1' or self.exponent == 1:
            return ''
        if not style:
            return f"^{self.exponent}"
        if 'tex' in style.lower():
            return f"^{{{self.exponent}}}"
        raise ValueError(f"Can't format {self.exponent}")

    def __int__(self):
        """Called for int(self)."""
        return self._cast_to(int)

    def __float__(self):
        """Called for float(self)."""
        return self._cast_to(float)

    _T = typing.TypeVar('_T', int, float)
    def _cast_to(self, __type: _T) -> _T:
        """Internal method for casting to numeric type."""
        if self.base == '1':
            return __type(self.coefficient)
        errmsg = f"Can't convert term with base {self.base!r} to {__type}"
        raise TypeError(errmsg) from None

    def __eq__(self, other) -> bool:
        if isinstance(other, numbers.Real):
            return float(self) == float(other)
        if isinstance(other, str):
            term = OperandFactory().create(other)
            return super().__eq__(term)
        return super().__eq__(other)

    def __hash__(self) -> int:
        # NOTE: If we decide to implement `Operand.__hash__`, we will still need
        # to explicitly define `Term.__hash__`. See
        # https://docs.python.org/3/reference/datamodel.html#object.__hash__ for
        # an explanation.
        return hash(self.attrs)


def asterms(these: typing.Iterable[str]):
    """Convert strings to terms, if possible."""
    return [OperandFactory().create(this) for this in these]


class OperandTypeError(TypeError):
    pass


class OperandValueError(ValueError):
    pass


T = typing.TypeVar('T', bound=Part)
class PartMatch(typing.Generic[T], iterables.ReprStrMixin):
    """An object that represents the result of a RE pattern match."""

    def __new__(cls, result, context):
        """Check argument types."""
        if not isinstance(result, Part):
            raise TypeError(
                f"Result must be an Part"
                f", not {type(result)}"
            ) from None
        if not isinstance(context, (re.Match, typing.Mapping)):
            raise TypeError(
                f"Context may be a Match object or a Mapping"
                f", not {type(context)}"
            )
        return super().__new__(cls)

    def __init__(
        self,
        result: T,
        context: typing.Union[re.Match, typing.Mapping],
    ) -> None:
        self.result = result
        """The result of the match attempt."""
        self._context = self._set_context(context)

    def _set_context(self, user: typing.Any) -> dict:
        """Normalize the instance context from user input."""
        if isinstance(user, re.Match):
            return self._set_from_match(user)
        if isinstance(user, typing.Mapping):
            return self._set_from_mapping(user)

    def _set_from_match(self, match: re.Match):
        """Set the instance context from a match object."""
        return {
            'start': match.start(),
            'end': match.end(),
            'string': match.string,
        }

    def _set_from_mapping(self, mapping: typing.Mapping):
        """Set the instance context from a mapping."""
        attrs = (
            'start',
            'end',
            'string',
        )
        return {k: mapping.get(k) for k in attrs}

    @property
    def start(self) -> int:
        """The starting index in `string` of the match."""
        return self._context['start']

    @property
    def end(self) -> int:
        """The ending index in `string` of the match."""
        return self._context['end']

    @property
    def string(self) -> str:
        """The target string."""
        return self._context['string']

    @property
    def remainder(self) -> str:
        """The unparsed portion of `string` after `end`."""
        return self.string[self.end:]

    def __bool__(self) -> bool:
        """Always true, like `re.Match`."""
        return True

    def __str__(self) -> str:
        """A simplified representation of this object."""
        attrs = {
            'result': 'result',
            'context': '_context',
        }
        return ', '.join(f"{k}: {getattr(self, v)}" for k, v in attrs.items())


@typing.runtime_checkable
class PartFactory(typing.Protocol):
    """Protocol for symbolic factories."""

    @abc.abstractmethod
    def parse(self) -> PartMatch:
        pass


class OperatorFactory(PartFactory):
    """A factory that produces symbolic operators."""

    def __init__(
        self,
        multiply: str='*',
        divide: str='/',
    ) -> None:
        mul = fr'\{multiply}'
        div = fr'\{divide}'
        self.patterns = {
            'multiply': re.compile(
                fr'(?<!{div})(\s*{mul}\s*)(?!{div})'
            ),
            'divide': re.compile(
                fr'(?<!{mul})(\s*{div}\s*)(?!{mul})'
            ),
            'sqrt': re.compile(r'\s*sqrt\s*')
        }
        """Compiled regular expressions for symbolic operators."""

    def parse(self, string: str):
        """Extract an operator at the start of `string`, possible."""
        for key in self.patterns:
            if match := self.patterns[key].match(string):
                return PartMatch(
                    result=Operator(key),
                    context=match,
                )


class OperandFactory(PartFactory):
    """A factory that produces symbolic operands."""

    rational = r""" # Modeled after `fractions._RATIONAL_FORMAT`
        [-+]?                 # an optional sign, ...
        (?=\d|\.\d)           # ... only if followed by <digit> or .<digit>
        \d*                   # and a possibly empty numerator
        (?:                   # followed by ...
            (?:/\d+?)         # ... an optional denominator
        |                     # OR
            (?:\.\d*)?        # ... an optional fractional part,
            (?:[eE][-+]?\d+)? #     and an optional exponent
        )
    """
    base = r"""
        [a-zA-Z#_]+ # one or more accepted non-digit character(s)
        \d*         # followed by optional digits
    """

    def __init__(
        self,
        opening: str='(',
        closing: str=')',
        raising: str='^',
    ) -> None:
        exponent = fr'\{raising}{self.rational}'
        self.patterns = {
            'constant': re.compile(
                fr'(?P<coefficient>{self.rational})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'variable': re.compile(
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>{self.base})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'complex': re.compile(
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>\{opening}.+?\{closing})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'exponent': re.compile(exponent, re.VERBOSE),
            'opening': re.compile(fr'\{opening}', re.VERBOSE),
            'closing': re.compile(fr'\{closing}', re.VERBOSE),
            'raising': re.compile(fr'\{raising}', re.VERBOSE)
        }
        """Compiled regular expressions for symbolic operands."""

    _argtypes = {
        'coefficient': numbers.Real,
        'base': str,
        'exponent': (numbers.Real, str),
    }
    @classmethod
    def isvalid(cls, name: str, this: typing.Any):
        """True if `this` is valid for use as the named attribute."""
        return isinstance(this, cls._argtypes[name])

    def normalize(self, *args):
        """Extract attributes from the given argument(s)."""
        try:
            nargs = len(args)
        except TypeError:
            raise OperandTypeError(args) from None
        if nargs == 1:
            return self._length_1(args)
        if nargs == 2:
            return self._length_2(args)
        if nargs == 3:
            return self._length_3(args)
        raise OperandValueError(
            f"{self.__class__.__qualname__}"
            f" accepts 1, 2, or 3 arguments"
            f" (got {nargs})"
        )

    def _length_1(self, args: typing.Any):
        """Normalize a length-1 argument tuple, if possible.

        A length-1 tuple may represent either:

        - coefficient <Real>
        - base <str>

        If it has one of these forms, this method will substitute the default
        value for the missing attributes; otherwise, it will raise an exception.
        """
        arg = args[0]
        names = ('base', 'coefficient')
        for name in names:
            given = {name: arg}
            if self.isvalid(name, arg):
                return self.standardize(fill=True, **given)
        raise OperandTypeError(
            "A single argument may be either"
            " a coefficient <Real> or a base <str>;"
            f" not {type(arg)}"
        )

    def _length_2(self, args: typing.Any):
        """Normalize a length-2 argument tuple, if possible.

        A length-2 tuple may represent either:

        - (base <str>, exponent <Real or str>)
        - (coefficient <Real>, exponent <Real or str>)
        - (coefficient <Real>, base <str>)

        If it has one of these forms, this method will substitute the default
        value for the missing attribute; otherwise, it will raise an exception.
        """
        combinations = itertools.combinations(self._argtypes, 2)
        for names in combinations:
            given = dict(zip(names, args))
            if all(self.isvalid(name, arg) for name, arg in given.items()):
                return self.standardize(fill=True, **given)
        badtypes = [type(arg) for arg in args]
        raise OperandTypeError(
            "Acceptable two-argument forms are"
            " (base <str>, exponent <Real or str>),"
            " (coefficient <Real>, exponent <Real or str>),"
            " or"
            " (coefficient <Real>, base <str>);"
            " not"
            f"({', '.join(str(t) for t in badtypes)})"
        )

    def _length_3(self, args: typing.Any):
        """Normalize a length-3 argument tuple.

        A length-3 tuple must have the form (coefficient <Real>, base <str>,
        exponent <Real or str>).
        """
        return self.standardize(
            coefficient=args[0],
            base=args[1],
            exponent=args[2],
            fill=True,
        )

    def create(self, *args, strict: bool=False) -> typing.Optional[Operand]:
        """Create an operand from input.

        Parameters
        ----------
        *args
            The object(s) from which to create an operand, if possible. This may
            take one of the following forms: a single string representing the
            base operand; a numerical coefficient and a base string; a base
            string and a numerical exponent; or a coefficient, base, and
            exponent. A missing coefficient or exponent will default to 1.

        strict : bool, default=false
            If true, this method will return `None` if it is unable to create an
            operand from `*args`. The default behavior is to return the input
            (with default coefficient and exponent, if necessary) as an instance
            of `~algebra.Operand`.

        Returns
        -------
        `~algebra.Operand` or `None`
            An instance of `~algebra.Operand` or one of its subclasses if the
            input arguments represent a valid operand. The `strict` keyword
            dictates the return behavior when input does not produce an operand.

        Notes
        -----
        This method will create the most general symbolic operand possible from
        the initial string. It will parse a simple symbolic operand into a
        coefficient, variable, and exponent but it will not attempt to fully
        parse a general symbolic operand into simpler operands (i.e. symbolic
        terms). In other words, it will do as little work as possible to extract
        a coefficient and exponent, and the expression on which they operate. If
        all attempts to determine appropriate attributes fail, the value of
        `strict` controls its return behavior.
        
        See note at `parse` for differences between this method and that.

        The following examples use the general symbolic operands described in
        `~algebra.Operand` to illustrate the minimal parsing described above:

        - `'a * b^2'` <=> `'(a * b^2)'` <=> `'(a * b^2)^1'` -> `1, 'a * b^2', 1`
        - `'2a * b^2'` -> `1, '2a * b^2', 1`
        - `'2(a * b^2)'` -> `2, 'a * b^2', 1`
        - `'(a * b^2)^3'` -> `1, 'a * b^2', 3`
        - `'2(a * b^2)^3'` -> `2, 'a * b^2', 3`
        - `'(a * b^2)^3/2'` -> `1, 'a * b^2', '3/2'`
        - `'((a / b^2)^3 * c)^2'` -> `1, '(a / b^2)^3 * c', 2`
        - `'(a / b^2)^3 * c^2'` -> `1, '(a / b^2)^3 * c^2', 1`
        """
        c0, b0, e0 = self.normalize(*args).values()
        ends = (b0[0], b0[-1])
        if any(self.patterns['raising'].match(c) for c in ends):
            raise OperandValueError(b0) from None
        match = self.search(b0, mode='fullmatch')
        if not match:
            if not strict:
                return Operand(c0, b0, e0)
            return
        if not isinstance(match.result, Operand):
            raise TypeError(
                f"Expected Operand but got {type(match.result)}"
            ) from None
        c1, base, e1 = match.result.attrs
        coefficient = c0 * (c1 ** e0)
        exponent = e1 * e0
        if not isinstance(match.result, Term):
            interior = self.create(base)
            if isinstance(interior, Term):
                coefficient *= interior.coefficient ** exponent
                base = interior.base
                exponent *= interior.exponent
                return Term(
                    coefficient=coefficient,
                    base=base,
                    exponent=exponent,
                )
            return Operand(
                coefficient=coefficient,
                base=base,
                exponent=exponent,
            )
        return Term(
            coefficient=coefficient,
            base=base,
            exponent=exponent,
        )

    def parse(self, string: str):
        """Extract an operand at the start of `string`, possible.
        
        Notes
        -----
        The primary difference between `~create` and `~parse` is as follows:

        - `~create` resolves input into (coefficient, base, exponent), then
          creates an appropriate operand from the base string, then applies the
          input coefficient and exponent, and finally returns the operand.

        - `~parse` attempts to match an operand at the start of a string, then
          creates an appropriate operand from only that substring, and finally
          returns the operand and the remainder of the string.
        """
        stripped = string.strip()
        if match := self.search(stripped):
            return match

    def search(self, string: str, **kwargs):
        """Search for an operand in the given string.
        
        Parameters
        ----------
        string
            The string to which to apply pattern-matching methods in an attempt
            to find an appropriate operand.

        **kwargs
            Keyword arguments to pass to the pattern-matching methods.

        Returns
        -------
        `~symbolic.PartMatch` or `None`
            An object representing the matched substring and contextual
            information, if any attemp to match was successful. If no method
            found an operand in `string` (subject to any given keyword
            arguments), this method will return `None`.
        """
        methods = (
            self._match_simplex,
            self._match_complex,
        )
        if match := iterables.apply(methods, string, **kwargs):
            return match

    def _match_simplex(
        self,
        string: str,
        mode: str='match',
        start: int=0,
    ) -> typing.Optional[PartMatch[Operand]]:
        """Attempt to find an irreducible term at the start of `string`.

        Notes
        -----
        This method tries to match the 'variable' pattern before the 'constant'
        pattern because `re.match` will find a match for 'constant' at the start
        of any variable term with an explicit coefficient.
        """
        target = string[start:]
        matches = {
            key: self._get_match_method(key, mode)(target)
            for key in ('variable', 'constant')
        }
        if not any(matches.values()):
            return
        if all(matches.values()):
            same = matches['variable'][0] == matches['constant'][0]
            key = 'constant' if same else 'variable'
            build_method = self._get_build_method(key)
            return build_method(matches[key])
        for key, match in matches.items():
            if match:
                return self._get_build_method(key)(match)

    def _get_match_method(
        self,
        pattern: str,
        mode: str,
    ) -> typing.Callable[[str], re.Match]:
        """Look up the appropriate matching method for `pattern` and `mode`."""
        return getattr(self.patterns[pattern], mode)

    def _get_build_method(
        self,
        pattern: str,
    ) -> typing.Callable[[re.Match], PartMatch[Operand]]:
        """Look up the appropriate building method for `pattern`."""
        return getattr(self, f'_build_{pattern}')

    def _build_variable(self, match: re.Match):
        """Build a variable term from a match object."""
        standard = self.standardize(**match.groupdict(), fill=True)
        return PartMatch(
            result=Term(**standard),
            context=match,
        )

    def _build_constant(self, match: re.Match):
        """Build a constant term from a match object."""
        standard = self.standardize(**match.groupdict(), fill=True)
        coefficient = standard['coefficient'] ** standard['exponent']
        return PartMatch(
            result=Term(coefficient=float(coefficient)),
            context=match,
        )

    def _match_complex(
        self,
        string: str,
        mode: str='match',
        start: int=0,
    ) -> typing.Optional[PartMatch[Operand]]:
        """Attempt to match a complex operand at the start of `string`."""
        target = string[start:]
        bounds = self.find_bounds(target)
        if not bounds:
            return
        i0, end = bounds
        result = {'base': target[i0+1:end-1]}
        if match := self._match_simplex(target[:i0], mode='fullmatch'):
            result['coefficient'] = match.result.coefficient
            i0 = 0
        if mode == 'match' and i0 != 0:
            return
        if exp := self.patterns['exponent'].match(target[end:]):
            result['exponent'] = exp[0]
            end += exp.end()
        if mode == 'fullmatch' and (i0, end) != (0, len(target)):
            return
        standard = self.standardize(**result, fill=True)
        return PartMatch(
            result=Operand(**standard),
            context={'end': end, 'start': start, 'string': string},
        )

    def find_bounds(self, string: str):
        """Find the indices of the first bounded substring, if any.
        
        A bounded substring is any portion of `string` that is bounded on the
        left by the opening separator and on the right by the closing separator.
        Opening and closing separators are an immutable attribute of an instance
        of this class.

        Parameters
        ----------
        string
            The string in which to search for a bounded substring.

        Returns
        -------
        tuple of int, or `None`
            The index of the leftmost opening separator and the index of the
            first character beyond the rightmost closing separator (possibly the
            end), if there is a bounded substring; otherwise, `None`. The
            convention is such that if `start, end = find_bounds(string)`,
            `string[start:end]` will produce the bounded substring with bounds.

        Examples
        --------
        Define a list of test strings::

            >>> strings = [
            ...     '(a*b)',
            ...     '(a*b)^2',
            ...     '3(a*b)',
            ...     '3(a*b)^2',
            ...     '3(a*b)^2 * (c*d)',
            ...     '4a^4',
            ...     '3(4a^4)^3',
            ...     '2(3(4a^4)^3)^2',
            ... ]

        Create an instance of this class with the default operators and
        separators::

            >>> operand = algebra.OperandFactory()

        Find the bounding indices of each test string, if any, and display the
        corresponding substring::

            >>> for string in strings:
            ...     bounds = operand.find_bounds(string)
            ...     print(f"{bounds}: {string!r} -> ", end='')
            ...     if bounds:
            ...         start, end = bounds
            ...         result = f"{string[start:end]}"
            ...     else:
            ...         result = string
            ...     print(f"{result!r}")
            ... 
            (0, 5): '(a*b)' -> '(a*b)'
            (0, 5): '(a*b)^2' -> '(a*b)'
            (1, 6): '3(a*b)' -> '(a*b)'
            (1, 6): '3(a*b)^2' -> '(a*b)'
            (1, 6): '3(a*b)^2 * (c*d)' -> '(a*b)'
            None: '4a^4' -> '4a^4'
            (1, 7): '3(4a^4)^3' -> '(4a^4)'
            (1, 12): '2(3(4a^4)^3)^2' -> '(3(4a^4)^3)'
        """
        initialized = False
        count = 0
        i0 = 0
        for i, c in enumerate(string):
            if self.patterns['opening'].match(c):
                count += 1
                if not initialized:
                    i0 = i
                    initialized = True
            elif self.patterns['closing'].match(c):
                count -= 1
            if initialized and count == 0:
                return i0, i+1

    def standardize(
        self,
        fill: bool=False,
        **given
    ) -> typing.Dict[str, typing.Union[float, int, str, fractions.Fraction]]:
        """Cast to appropriate types and fill in defaults, if necessary."""
        full = {
            'coefficient': {'callable': self._standard_coefficient},
            'base': {'callable': self._standard_base},
            'exponent': {'callable': self._standard_exponent},
        }
        default = self.fill_defaults(**dict.fromkeys(full.keys()))
        updatable = full.copy() if fill else {k: full[k] for k in given}
        return {
            key: attr['callable'](given.get(key) or default[key])
            for key, attr in updatable.items()
        }

    def _standard_coefficient(self, v):
        """Convert input to a standard coefficient."""
        return fractions.Fraction(v or 1)

    def _standard_base(self, v):
        """Convert input to a standard base."""
        return str(v)

    def _standard_exponent(self, v):
        """Convert input to a standard exponent."""
        if isinstance(v, str):
            v = self.patterns['raising'].sub('', v)
        return fractions.Fraction(v or 1)

    def fill_defaults(self, **given):
        """Return the default value for any explicitly null arguments.

        If the given key-value pairs contain an argument with a null value
        (e.g., `None`), this method will replace it with the default value. It
        will pass all other values through unaltered and will not fill in
        default values corresponding to other keys.
        """
        defaults = {
            'coefficient': 1,
            'base': '1',
            'exponent': 1,
        }
        given.update(
            {
                key: defaults[key]
                for key, value in given.items()
                if key in defaults and not value
            }
        )
        return given

    def strip_separators(self, string: str):
        """Remove one opening and one closing separator."""
        opened = self.patterns['opening'].match(string[0])
        closed = self.patterns['closing'].match(string[-1])
        if not (opened and closed):
            return string
        string = self.patterns['opening'].sub('', string, count=1)
        string = self.patterns['closing'].sub('', string[::-1], count=1)
        return string[::-1]


class ParsingError(Exception):
    """Base class for exceptions encountered during symbolic parsing."""

    def __init__(self, arg: typing.Any) -> None:
        self.arg = arg


class RatioError(ParsingError):
    """The string contains multiple '/' on a single level."""

    def __str__(self) -> str:
        return (
            f"The expression '{self.arg}' contains ambiguous '/'."
            f" Please refer to the NIST guidelines"
            f" (https://physics.nist.gov/cuu/Units/checklist.html)"
            f" for more information."
        )


class ProductError(ParsingError):
    """The string contains a '*' after a '/'."""

    def __str__(self) -> str:
        return (
            f"The expression '{self.arg}' contains an ambiguous '*'."
            f" Please group '*' in parentheses when following '/'."
        )


class ParsingValueError(ValueError):
    """Cannot create an expression from the given string."""
    pass


class Iteration(iterables.ReprStrMixin):
    """An object that keeps track of parsing attributes."""

    __slots__ = ('string', 'operator', 'operand')

    def __init__(
        self,
        string: str,
        operator: Operator=None,
        operand: Operand=None,
    ) -> None:
        self.string = string
        self.operator = operator
        self.operand = operand

    @property
    def _attrs(self):
        """Internal mapping of current attribute values."""
        return {name: getattr(self, name) for name in self.__slots__}

    def copy(self):
        """Make a copy of this instance."""
        return type(self)(**self._attrs)

    def __str__(self):
        """A simplified representation of this object."""
        return ', '.join(f"{k}={v!r}" for k, v in self._attrs.items())


class Parser:
    """A tool for parsing symbolic expressions."""

    def __init__(
        self,
        multiply: str='*',
        divide: str='/',
        opening: str='(',
        closing: str=')',
        raising: str='^',
        operator_order: str='ignore',
    ) -> None:
        """
        Initialize a parser with fixed tokens.

        Parameters
        ----------
        multiply : string, default='*'
            The token that represents multiplication.

        divide : string, default='/'
            The token that represents division.

        opening : string, default='('
            The token that represents an opening separator.

        closing : string, default='('
            The token that represents a closing separator.

        raising : string, default='^'
            The token that represents raising to a power (exponentiation).

        operator_order : {'ignore', 'error'}
            Determines how the parser responds when operator order violates NIST
            guidelines. If set to `'ignore'` (default), it will treat operators
            independent of one another. If set to `'error'`, the parser will
            raise an exception based on the type of violation.
        """
        self.operands = OperandFactory(opening, closing, raising)
        self.operators = OperatorFactory(multiply, divide)
        self.parsers = (self.operands, self.operators)
        self.tokens = {
            'multiply': multiply,
            'divide': divide,
            'opening': opening,
            'closing': closing,
            'raising': raising,
        }
        self._operator_order = operator_order

    def parse(self, string: str):
        """Resolve the given string into individual terms."""
        operand = Operand(base=string)
        return self._resolve_operations(operand)

    def _resolve_operations(self, current: Operand) -> typing.List[Term]:
        """Separate a symbolic group into operators and operands."""
        operands = self._parse_operand(current)
        return [
            term for operand in operands
            for term in self._update_terms(operand)
        ] + [Term(coefficient=current.coefficient)]

    def _parse_operand(self, initial: Operand) -> typing.List[Operand]:
        """Resolve a general operand into simpler operands.

        This method parses known operators and operands from the initial operand
        while preserving nested groups in the latter. Calling code may then pass
        those nested groups back in for further parsing.
        """
        operands = []
        current = Iteration(initial.base)
        previous = current.copy()
        while current.string:
            current = self._get_operator(initial, current, previous)
            current = self._get_operand(initial, current)
            if new := self._compute_operand(current):
                operands.append(new)
            previous = current.copy()
            current = Iteration(previous.string)
        return operands

    def _get_operator(
        self,
        initial: Operand,
        current: Iteration,
        previous: Iteration,
    ) -> Iteration:
        """Attempt to parse an operator from the current string."""
        if parsed := self.operators.parse(current.string):
            current.operator = parsed.result
            if exception := self._operator_error(
                    current.operator,
                    previous.operator,
                ): raise exception(initial)
            current.string = parsed.remainder
        return current

    def _get_operand(
        self,
        initial: Operand,
        current: Iteration,
    ) -> Iteration:
        """Attempt to parse an operand from the current string."""
        if parsed := self.operands.parse(current.string):
            current.operand = parsed.result ** initial.exponent
            current.string = parsed.remainder
        return current

    def _compute_operand(self, current: Iteration):
        """Create a new operand from the current iteration."""
        if current.operand and current.operator:
            return self._evaluate(current.operator, current.operand)
        if current.operand:
            return current.operand
        if current.operator:
            raise ParsingValueError("Operator without operand")
        raise ParsingValueError("Failed to parse string")

    def _operator_error(self, current: Operator, previous: Operator):
        """Check for known operator-related errors.
        
        This method checks for the following errors and returns the appropriate
        exception class if it finds one:

        - Multiple divisions on a single level (e.g., `'a / b / c'`), which
          results in a `RatioError`.

        - Multiplication after division on the same level (e.g., `'a / b * c'`),
          which results in a `ProductError`.

        Both of the examples shown above result in errors because they each
        introduce an ambiguous order of operations. Users can resolve the
        ambiguity by properly grouping terms in the expression. Continuing with
        the above examples, `'a / b / c'` should become `'(a / b) / c'` or `'a /
        (b / c)'`, and `'a / b * c'` should become `'(a / b) * c'` or `'a / (b *
        c)'`.
        """
        if self._operator_order == 'ignore':
            return
        if previous == 'divide':
            if current == 'divide':
                return RatioError
            if current == 'multiply':
                return ProductError

    def _evaluate(self, operator: Operator, operand: Operand):
        """Compute the effect of `operator` on `operand`."""
        if operator in {'multiply', 'identity'}:
            return operand
        if operator == 'divide':
            return operand ** -1
        if operator == 'sqrt':
            return operand ** 0.5
        raise ValueError(f"Unrecognized operator {operator!r}")

    def _update_terms(self, operand: Operand):
        """Store a new term or initiate further parsing."""
        # TODO: Consider extracting all coefficients, at least as separate
        # constant terms.
        if isinstance(operand, Term):
            return [operand]
        return self._resolve_operations(operand)


class OperandError(TypeError):
    pass


def standard(this, missing: str='1', joiner: str='*') -> str:
    """Convert `this` to a standard format.
    
    Parameters
    ----------
    this : string or iterable
        The object to convert.

    missing : string, default='1'
        The value to use if `this` is null.

    joiner : string, default='*'
        The string token to use when joining parts of an iterable argument.

    See Also
    --------
    `~symbolic.Expression`: A class that represents one or more terms joined by
    symbolic operators and grouped by separator characters. Instances support
    multiplication and division with strings or other instances, and
    exponentiation by real numbers. Instantiation automatically calls this
    function.
    """
    if not this:
        return missing
    if isinstance(this, str):
        return this
    try:
        iter(this)
    except TypeError:
        return str(this)
    else:
        return joiner.join(f"({part})" for part in this)


Instance = typing.TypeVar('Instance', bound='Expression')


class Expression(collections.abc.Sequence, iterables.ReprStrMixin):
    """An object representing a symbolic expression.

    If this class is instantiated with an existing instance, the result will be
    the same instance. Otherwise, it will create a new instance of this class
    from the given string or iterable after replacing operators and separators
    with their standard versions, if necessary.
    """

    def __new__(cls, *args, **kwargs):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, cls):
                if not kwargs:
                    return arg
                raise TypeError(
                    "Can't change parsing options on an existing expression"
                ) from None
        return super().__new__(cls)

    @typing.overload
    def __init__(self, expression: str, /, **kwargs) -> None:
        """Create an expression from a single string.
        
        Parameters
        ----------
        expression : string
            A single string to convert into an expression.

        **kwargs
            Keywords to pass to `~algebra.Parser`.

        Examples
        --------
        Create a symbolic expression from a string that represents the result of
        multiplying `a^3/2` by `b`, dividing by `c^1/2`, and squaring the ratio:

        >>> symbolic.Expression('(a^3/2 * b / c^1/2)^2')
        core.symbolic.Expression(a^3 b^2 c^-1)
        """

    @typing.overload
    def __init__(self, expression: typing.Iterable, /, **kwargs) -> None:
        """Create an expression from an iterable of any type.
        
        Parameters
        ----------
        expression : iterable
            An iterable of any type to initialize the new instance. All members
            must support conversion to a string.

        **kwargs
            Keywords to pass to `~algebra.Parser`.

        Examples
        --------
        Create a symbolic expression from a list of the individual string terms
        in the result of multiplying `a^3/2` by `b`, dividing by `c^1/2`, and
        squaring the ratio:

        >>> symbolic.Expression(['a^3', 'b', 'c^-1'])
        core.symbolic.Expression(a^3 b c^-1)
        """

    @typing.overload
    def __init__(self: Instance, expression: Instance, /) -> None:
        """Create an expression from an expression.

        This mode exists to support algorithms that don't know the type of
        argument until runtime. If the type is known, it is simpler to use the
        existing instance.

        Parameters
        ----------
        expression : `~symbolic.Expression`
            An existing instance of this class.

        Examples
        --------
        Create an instance from a string:

        >>> this = symbolic.Expression('a * b / c')

        Pass the first instance to this class:

        >>> that = symbolic.Expression(this)

        Both `this` and `that` represent the same expression... >>> this
        core.symbolic.Expression(a b c^-1) >>> that core.symbolic.Expression(a b
        c^-1)

        ...because they are the same object. >>> that is this True
        """

    def __init__(self, expression, **kwargs) -> None:
        string = standard(expression, joiner=' * ')
        terms = Parser(**kwargs).parse(string)
        self.terms = reduce(terms)
        """The symbolic terms in this expression."""

    def __iter__(self) -> typing.Iterator[Term]:
        return iter(self.terms)

    def __len__(self) -> int:
        return len(self.terms)

    @typing.overload
    def __getitem__(self, index: typing.SupportsIndex) -> Term: ...

    @typing.overload
    def __getitem__(self, index: slice) -> typing.Iterable[Term]: ...

    def __getitem__(self, index):
        """Access terms via standard indexing."""
        if isinstance(index, typing.SupportsIndex):
            idx = int(index)
            if idx > len(self):
                raise IndexError(index)
            if idx < 0:
                idx += len(self)
            return self.terms[idx]
        return self.terms[index]

    def __str__(self) -> str:
        """A simplified representation of this instance."""
        return self.format()

    def format(self, separator: str=' ', style: str=None):
        """Join symbolic terms into a string."""
        formatted = (term.format(style=style) for term in self)
        return separator.join(formatted)

    def difference(self, other, symmetric: bool=False, split: bool=False):
        """Compute the difference between two expressions.
        
        Parameters
        ----------
        other
            The object with respect to which to compute the difference. If
            `other` is not a `~symbolic.Expression`, this method will convert it
            to one before proceeding.

        symmetric : bool, default=False
            If true, compute the symmetric difference between this expression
            and `other`.

        split : bool, default=False
            If true, return the one-sided differences in a ``list``. The first
            element contains the terms in this expression that are not in
            `other`, and the second element contains the terms in `other` that
            are not in this expression.

        Notes
        -----
        The `split` keyword argument takes precedence over the `symmetric`
        keyword argument because the result of the former contains more
        imformation than the result of the latter. See Examples for a suggestion
        on converting a split result into a symmetric result.

        Examples
        --------
        Consider the following two expressions:
        
        >>> e0 = symbolic.Expression('a * b')
        >>> e1 = symbolic.Expression('a * c')

        Their formal (one-sided) difference is

        >>> e0.difference(e1)
        {core.symbolic.Term(b)}

        Their formal symmetric difference is

        >>> e0.difference(e1, symmetric=True)
        {core.symbolic.Term(b), core.symbolic.Term(c)}

        Passing ``split=True`` produces a ``list`` of ``set``s

        >>> e0.difference(e1, split=True)
        [{core.symbolic.Term(b)}, {core.symbolic.Term(c)}]

        To convert a split result into a symmetric result, simply compute the
        union of the former:

        >>> symmetric = e0.difference(e1, symmetric=True)
        >>> split = e0.difference(e1, split=True)
        >>> set.union(*split) == symmetric
        True
        """
        if not isinstance(other, Expression):
            other = type(self)(other)
        s0 = set(self.terms)
        s1 = set(other.terms)
        if split:
            return [s0 - s1, s1 - s0]
        if symmetric:
            return s0 ^ s1
        return s0 - s1

    def __hash__(self):
        """Compute hash(self)."""
        return hash(tuple(self.terms))

    def __eq__(self, other) -> bool:
        """True if two expressions have the same symbolic terms.

        This method defines two expressions as equal if they have equivalent
        lists of symbolic terms (a.k.a simple parts), regardless of order, after
        parsing. Two expressions with different numbers of terms are always
        false. If the expressions have the same number of terms, this method
        will sort the triples (first by base, then by exponent, and finally by
        coefficient) and compare the sorted lists. Two expressions are equal if
        and only if their sorted lists of terms are equal.

        If `other` is not an instance of this class, this method will first
        attempt to convert it.
        """
        if not isinstance(other, Expression):
            other = type(self)(other)
        if len(self) != len(other):
            return False
        key = attrgetter('base', 'exponent', 'coefficient')
        return sorted(self, key=key) == sorted(other, key=key)

    def __mul__(self, other):
        """Called for self * other.

        This method implements multiplication between two expressions by
        reducing the exponents of terms with a common base. If `other` is a
        string, it will first attempt to convert it to an `Expression`.
        """
        if not isinstance(other, Expression):
            other = type(self)(other)
        if not other:
            return NotImplemented
        reduced = reduce(self, other)
        return type(self)(reduced)

    def __rmul__(self, other: typing.Any):
        """Called for other * self."""
        return type(self)(other).__mul__(self)

    def __truediv__(self, other):
        """Called for self / other.

        This method implements division between two expressions by raising all
        terms in `other` to -1, then reducing the exponents of terms with a
        common base. If `other` is a string, it will first attempt to convert it
        to an `Expression`.
        """
        if not isinstance(other, Expression):
            other = type(self)(other)
        if not other:
            return NotImplemented
        return type(self)(reduce(self, [term ** -1 for term in other]))

    def __rtruediv__(self, other: typing.Any):
        """Called for other / self."""
        return type(self)(other).__truediv__(self)

    def __pow__(self, exp: numbers.Real):
        """Called for self ** exp.

        This method implements exponentiation of an expression by raising all
        terms to the given power, then reducing exponents of terms with a common
        base. It will first attempt to convert `exp` to a float.
        """
        exp = float(exp)
        if not exp:
            return NotImplemented
        terms = [pow(term, exp) for term in self]
        return type(self)(reduce(terms))

    def apply(self: Instance, update: typing.Callable) -> Instance:
        """Create a new expression by applying the given callable object.
        
        Parameters
        ----------
        update : callable
            The callable object that this method should use to update the base
            of each term in this expression.

        Returns
        -------
        `~symbolic.Expression`
        """
        bases = [update(term.base) for term in self]
        exponents = [term.exponent for term in self]
        result = bases[0] ** exponents[0]
        for base, exponent in zip(bases[1:], exponents[1:]):
            result *= base ** exponent
        return result


algebraic.Multiplicative.register(Expression)


Expressable = typing.TypeVar(
    'Expressable',
    str,
    typing.Iterable,
    Expression,
)
Expressable = typing.Union[str, typing.Iterable, Expression]


def reduce(*groups: typing.Iterable[Term]):
    """Algebraically reduce terms with equal bases.

    Parameters
    ----------
    *groups : tuple of iterables
        One or more iterables of `~algebra.Term` instances. If there are
        multiple groups, this method will combine all terms it finds in the full
        collection of groups.

    Notes
    -----
    This function will sort terms in order of ascending exponent, and
    alphabetically for equal exponents.
    """
    terms = [term for group in groups for term in group]
    reduced = {}
    for term in terms:
        if term.base in reduced:
            reduced[term.base]['coefficient'] *= term.coefficient
            reduced[term.base]['exponent'] += term.exponent
        else:
            attributes = {
                'coefficient': term.coefficient,
                'exponent': term.exponent,
            }
            reduced[term.base] = attributes
    fracs = [
        fractions.Fraction(v['coefficient'])
        for v in reduced.values()
    ]
    tmp = [
        Term(base=k, exponent=v['exponent'])
        for k, v in reduced.items()
        if k != '1' and v['exponent'] != 0
    ]
    # Sort: high to low in exponent, followed by alphabetic in base.
    variables = sorted(
        sorted(tmp, key=attrgetter('base')),
        key=attrgetter('exponent'),
        reverse=True,
    )
    c = functools.reduce(lambda x, y: x*y, fracs)
    constant = [Term(coefficient=c)]
    if not variables:
        return constant
    if c == 1:
        return variables
    return constant + variables


def equality(a, b) -> Expression:
    """Symbolically compute a == b."""
    x, y = (Expression(i) for i in (a, b))
    return x == y


def product(a, b) -> Expression:
    """Symbolically compute a * b."""
    x, y = (Expression(i) for i in (a, b))
    return x * y


def ratio(a, b) -> Expression:
    """Symbolically compute a / b."""
    x, y = (Expression(i) for i in (a, b))
    return x / y


def power(a, n) -> Expression:
    """Symbolically compute a ** n."""
    return Expression(a) ** n




import abc
import math
import numbers
import operator as standard
import typing

import numpy


@typing.runtime_checkable
class Orderable(typing.Protocol):
    """Protocol for objects that support ordering.
    
    Instance checks against this ABC will return `True` iff the instance
    implements the following methods: `__lt__`, `__gt__`, `__le__`, `__ge__`,
    `__eq__`, and `__ne__`. It exists to support type-checking orderable objects
    outside the `~algebraic.Quantity` framework (e.g., pure numbers).
    """

    __slots__ = ()

    @abc.abstractmethod
    def __lt__(self, other):
        pass

    @abc.abstractmethod
    def __eq__(self, other):
        pass

    @abc.abstractmethod
    def __le__(self, other):
        pass

    @abc.abstractmethod
    def __gt__(self, other):
        pass

    @abc.abstractmethod
    def __ge__(self, other):
        pass

    @abc.abstractmethod
    def __ne__(self, other):
        pass


class Ordered(abc.ABC):
    """Abstract base class for all objects that support relative ordering.

    Concrete implementations of this class must define the six binary comparison
    operators (a.k.a "rich comparison" operators): `__lt__`, `__gt__`, `__le__`,
    `__ge__`, `__eq__`, and `__ne__`.

    The following default implementations are available by calling their
    equivalents on `super()`:

    - `__ne__`: defined as not equal.
    - `__le__`: defined as less than or equal.
    - `__gt__`: defined as not less than and not equal.
    - `__ge__`: defined as not less than.
    """

    __slots__ = ()

    __hash__ = None

    @abc.abstractmethod
    def __lt__(self, other) -> bool:
        """True if self < other."""
        pass

    @abc.abstractmethod
    def __eq__(self, other) -> bool:
        """True if self == other."""
        pass

    @abc.abstractmethod
    def __le__(self, other) -> bool:
        """True if self <= other."""
        return self.__lt__(other) or self.__eq__(other)

    @abc.abstractmethod
    def __gt__(self, other) -> bool:
        """True if self > other."""
        return not self.__le__(other)

    @abc.abstractmethod
    def __ge__(self, other) -> bool:
        """True if self >= other."""
        return not self.__lt__(other)

    @abc.abstractmethod
    def __ne__(self, other) -> bool:
        """True if self != other."""
        return not self.__eq__(other)


Self = typing.TypeVar('Self', bound='Additive')


class Additive(abc.ABC):
    """Abstract base class for additive objects."""

    __slots__ = ()

    @abc.abstractmethod
    def __add__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __radd__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __sub__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __rsub__(self: Self, other) -> Self:
        pass


Self = typing.TypeVar('Self', bound='Multiplicative')


class Multiplicative(abc.ABC):
    """Abstract base class for multiplicative objects."""

    __slots__ = ()

    @abc.abstractmethod
    def __mul__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __rmul__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __truediv__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __rtruediv__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __pow__(self: Self, other) -> Self:
        pass

    @abc.abstractmethod
    def __rpow__(self: Self, other) -> Self:
        pass


class Base(Ordered, Additive, Multiplicative):
    """Base class for algebraic quantities.

    Concrete subclasses of this class must implement the six comparison
    operators,
        - `__lt__` (less than; called for `self < other`)
        - `__gt__` (greater than; called for `self > other`)
        - `__le__` (less than or equal to; called for `self <= other`)
        - `__ge__` (greater than or equal to; called for `self >= other`)
        - `__eq__` (equal to; called for `self == other`)
        - `__ne__` (not equal to; called for `self != other`)
    
    the following unary arithmetic operators,
        - `__abs__` (absolute value; called for `abs(self)`)
        - `__neg__` (negative value; called for `-self`)
        - `__pos__` (positive value; called for `+self`)

    and the following binary arithmetic operators,
        - `__add__` (addition; called for `self + other`)
        - `__radd__` (reflected addition; called for `other + self`)
        - `__sub__` (subtraction; called for `self - other`)
        - `__rsub__` (reflected subtraction; called for `other - self`)
        - `__mul__` (multiplication; called for `self * other`)
        - `__rmul__` (reflected multiplication; called for `other * self`)
        - `__truediv__` (division; called for `self / other`)
        - `__rtruediv__` (reflected division; called for `other / self`)
        - `__pow__` (exponentiation; called for `self ** other`)
        - `__rpow__` (reflected exponentiation; called for `other ** self`)

    Any required method may return `NotImplemented`.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __abs__(self):
        """Implements abs(self)."""
        pass

    @abc.abstractmethod
    def __neg__(self):
        """Called for -self."""
        pass

    @abc.abstractmethod
    def __pos__(self):
        """Called for +self."""
        pass


Self = typing.TypeVar('Self', bound='SupportsNeg')


@typing.runtime_checkable
class SupportsNeg(typing.Protocol):
    """Protocol for objects that support negation (``-self``)."""

    __slots__ = ()

    @abc.abstractmethod
    def __neg__(self: Self) -> Self:
        pass


class Real(Base):
    """Abstract base class for all real-valued objects.
    
    This class is similar to ``numbers.Real``, but it does not presume to
    represent a single value.
    
    Concrete subclasses of this object must implement all the `~algebraic.Base`
    operators except for `__sub__` and `__rsub__` (defined here with respect to
    `__neg__`). Subclasses may, of course, override these base implementations.
    """

    def __sub__(self, other: SupportsNeg):
        """Called for self - other."""
        return self + -other

    def __rsub__(self, other: SupportsNeg):
        """Called for other - self."""
        return -self + other


Real.register(numbers.Real)
Real.register(numpy.ndarray) # close enough for now...


class Quantity(abc.ABC):
    """ABC for algebraic quantities.
    
    Concrete subclasses must define the built-in `__eq__` method and an
    `implement` method that computes the result of a given operation on specific
    operands. Operator mixin classes can leverage this class to programmatically
    implement operator methods required by `~algebraic.Base` or similar ABCs.
    Concrete subclasses of this class are always true in a boolean sense.
    """

    def __bool__(self) -> bool:
        """Always true for a valid instance."""
        return True

    def __abs__(self):
        """Called for abs(self)."""
        return self.implement(abs, 'arithmetic')

    def __pos__(self):
        """Called for +self."""
        return self.implement(standard.pos, 'arithmetic')

    def __neg__(self):
        """Called for -self."""
        return self.implement(standard.neg, 'arithmetic')

    def __ne__(self, other) -> bool:
        """Called for self != other."""
        return not self == other

    def __lt__(self, other) -> bool:
        """Called for self < other."""
        return self.implement(standard.lt, 'comparison', other)

    def __le__(self, other) -> bool:
        """Called for self <= other."""
        return self.implement(standard.le, 'comparison', other)

    def __gt__(self, other) -> bool:
        """Called for self > other."""
        return self.implement(standard.gt, 'comparison', other)

    def __ge__(self, other) -> bool:
        """Called for self >= other."""
        return self.implement(standard.ge, 'comparison', other)

    def __add__(self, other):
        """Called for self + other."""
        return self.implement(standard.add, 'forward', other)

    def __radd__(self, other):
        """Called for other + self."""
        return self.implement(standard.add, 'reverse', other)

    def __iadd__(self, other):
        """Called for self += other."""
        return self.implement(standard.add, 'inplace', other)

    def __sub__(self, other):
        """Called for self - other."""
        return self.implement(standard.sub, 'forward', other)

    def __rsub__(self, other):
        """Called for other - self."""
        return self.implement(standard.sub, 'reverse', other)

    def __isub__(self, other):
        """Called for self -= other."""
        return self.implement(standard.sub, 'inplace', other)

    def __mul__(self, other):
        """Called for self * other."""
        return self.implement(standard.mul, 'forward', other)

    def __rmul__(self, other):
        """Called for other * self."""
        return self.implement(standard.mul, 'reverse', other)

    def __imul__(self, other):
        """Called for self *= other."""
        return self.implement(standard.mul, 'inplace', other)

    def __truediv__(self, other):
        """Called for self / other."""
        return self.implement(standard.truediv, 'forward', other)

    def __rtruediv__(self, other):
        """Called for other / self."""
        return self.implement(standard.truediv, 'reverse', other)

    def __itruediv__(self, other):
        """Called for self /= other."""
        return self.implement(standard.truediv, 'inplace', other)

    def __pow__(self, other):
        """Called for self ** other."""
        return self.implement(standard.pow, 'forward', other)

    def __rpow__(self, other):
        """Called for other ** self."""
        return self.implement(standard.pow, 'reverse', other)

    def __ipow__(self, other):
        """Called for self **= other."""
        return self.implement(standard.pow, 'inplace', other)

    @abc.abstractmethod
    def implement(self, func: typing.Callable, mode: str, *others, **kwargs):
        """Implement a standard operator."""
        pass


class Scalar(Quantity):
    """ABC for single-valued algebraic quantities.
    
    This class `~algebraic.Quantity` to define two numeric cast operators
    (`_int__` and `__float__`) and four unary arithmetic operators (`__round__`,
    `__floor__`, `__ceil__`, and `__trunc__`) in terms of the abstract method
    `implement`.
    """

    def __int__(self):
        """Called for int(self)."""
        return self.implement(int, 'cast')

    def __float__(self):
        """Called for float(self)."""
        return self.implement(float, 'cast')

    def __round__(self):
        """Called for round(self)."""
        return self.implement(round, 'arithmetic')

    def __floor__(self):
        """Called for math.floor(self)."""
        return self.implement(math.floor, 'arithmetic')

    def __ceil__(self):
        """Called for math.ceil(self)."""
        return self.implement(math.ceil, 'arithmetic')

    def __trunc__(self):
        """Called for math.trunc(self)."""
        return self.implement(math.trunc, 'arithmetic')





import collections
import collections.abc
import typing

from goats.core import aliased
from goats.core import datafile
from goats.core import index
from goats.core import iterables
from goats.core import metadata
from goats.core import metric
from goats.core import reference
from goats.core import variable


T = typing.TypeVar('T')


class Metadata(collections.UserDict):
    """Metadata attributes relevant to a dataset axis."""

    def __init__(
        self,
        name: typing.Union[str, typing.Iterable[str]]=None,
        unit: typing.Union[str, metric.Unit]=None,
    ) -> None:
        attrs = {}
        if name is not None:
            attrs['name'] = metadata.Name(name)
        if unit is not None:
            attrs['unit'] = metadata.Unit(unit)
        super().__init__(attrs)


class Indexer(iterables.ReprStrMixin):
    """An object that computes axis indices from user values."""

    def __init__(
        self,
        method: typing.Callable[..., index.Data],
        size: int,
    ) -> None:
        self._method = method
        """The method that converts target values into indices."""
        self.size = size
        """The maximum number of indices."""
        # NOTE: Instances of this class do not expect an array-like reference
        # object at initialization because knowledge of the appropriate
        # reference array is not always possible before calling `method`, and
        # defining reference-based attributes on only certain instances (or
        # subclasses) could create a misleading interface.

    def compute(self, *args, **kwargs):
        """Call the index-computing method."""
        return self._method(*args, **kwargs)

    def normalize(self, *user: T):
        """Convert user input into suitable target values."""
        if not user:
            return range(self.size)
        if isinstance(user[0], slice):
            return iterables.slice_to_range(user[0], stop=self.size)
        if isinstance(user[0], range):
            return user[0]
        return user

    def __str__(self) -> str:
        return f"{self._method.__qualname__}, size={self.size}"


Instance = typing.TypeVar('Instance', bound='Quantity')


class Quantity:
    """A callable representation of a dataset axis."""

    @typing.overload
    def __init__(self, __computer: Indexer, **meta) -> None:
        """Create a new axis from scratch."""

    @typing.overload
    def __init__(self: Instance, instance: Instance) -> None:
        """Create a new axis from an existing instance."""

    def __new__(cls, __a, **meta):
        """Return an existing instance, if applicable."""
        if not meta and isinstance(__a, cls):
            return __a
        return super().__new__(cls)

    def __init__(self, __a: Indexer, **meta) -> None:
        self._indexer = __a
        self._meta = meta
        self._reference = None

    @property
    def name(self) -> metadata.Name:
        """The name of this axis."""
        return self._meta.get('name')

    @property
    def unit(self) -> metadata.Unit:
        """The unit of this axis's values."""
        return self._meta.get('unit')

    @property
    def reference(self):
        """The full array of axis values."""
        if self._reference is None:
            self._reference = index.Array(self.index())
        return self._reference

    def index(self, *args, **kwargs):
        """Convert arguments into an index-like quantity."""
        targets = self._indexer.normalize(*args)
        unit = kwargs.pop('unit', self.unit)
        if unit:
            data = self._indexer.compute(targets, unit, **kwargs)
            meta = {**self._meta, 'unit': unit}
            return index.Quantity(data, **meta)
        data = self._indexer.compute(targets, **kwargs)
        return index.Quantity(data, **self._meta)

    # NOTE: The following unit-related logic includes significant overlap with
    # `measurable.Quantity`.
    def __getitem__(self, arg: typing.Union[str, metric.Unit]):
        """Set the unit of this object's values, if applicable.
        
        Notes
        -----
        Using this special method to change the unit supports a simple and
        relatively intuitive syntax but is arguably an abuse of notation.

        Raises
        ------
        TypeError
            User attempted to modify the unit of an unmeasurable quantity.

        ValueError
            The given unit is inconsistent with this quantity. Two units are
            mutually consistent if they have the same dimension in a known
            metric system.
        """
        if self.unit is None:
            raise TypeError(
                "Can't set the unit of an unmeasurable quantity."
            ) from None
        unit = (
            self.unit.norm[arg]
            if str(arg).lower() in metric.SYSTEMS else arg
        )
        if unit == self.unit:
            return self
        new = metadata.Unit(unit)
        if self.unit | new:
            return type(self)(self._indexer, unit=new, name=self.name)
        raise ValueError(
            f"The unit {str(unit)!r} is inconsistent with {str(self.unit)!r}"
        ) from None

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return ', '.join(f"{k}={str(v)!r}" for k, v in self._meta.items())


class IndexTypeError(Exception):
    """Invalid index argument."""


class Interface(aliased.MutableMapping):
    """An interface to the array axes available from a dataset."""

    def __init__(
        self,
        dataset: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        self.dataset = dataset
        """The dataset to which these axes belong."""
        self._variables = None
        self._system = metric.System(system or 'mks')
        indexers = {
            k: self.build_default(k)
            for k in self.dataset.axes.keys(aliased=True)
        }
        super().__init__(indexers)

    @property
    def variables(self):
        """The variable quantities that these axes support."""
        if self._variables is None:
            self._variables = variable.Interface(
                self.dataset,
                system=self.system,
            )
        return self._variables

    @property
    def system(self):
        """The metric system of these axes."""
        return self._system

    def __getitem__(self, __k: str):
        """Get the named axis object, if possible."""
        try:
            method = self._get_indexer(__k)
        except KeyError as err:
            raise KeyError(
                f"No known indexing method for {__k}"
            ) from err
        meta = self.get_metadata(__k)
        return Quantity(method, **meta)

    def _get_indexer(self, key: str):
        """Get the axis indexer for `key`, or use the default."""
        if key in self.keys():
            return super().__getitem__(key)
        raise KeyError(f"No indexing method for axis {key!r}")

    def build_default(self, key: str):
        """Define the default axis-indexer method."""
        n = self.dataset.axes[key].size
        def method(targets):
            try:
                indices = [int(arg) for arg in targets]
                if all(0 <= idx < n for idx in indices):
                    return index.Data(indices)
            except TypeError as err:
                raise IndexTypeError(
                    f"Can't convert {targets!r} to integer indices."
                ) from err
            raise ValueError(
                f"One or more index {targets} is outside the interval"
                " [0, {n-1}]"
            ) from None
        return Indexer(method, n)

    def get_metadata(self, key: str):
        """Get metadata attributes corresponding to `key`."""
        name = reference.NAMES.get(key, key)
        try:
            unit = self.variables[key].unit
        except KeyError:
            unit = None
        return Metadata(name=name, unit=unit)

    def resolve(
        self,
        names: typing.Iterable[str],
        mode: str='strict',
    ) -> typing.Tuple[str]:
        """Compute and order the available axes in `names`."""
        axes = self.dataset.available('axes').canonical
        ordered = tuple(name for name in axes if name in names)
        if mode == 'strict':
            return ordered
        extra = tuple(name for name in names if name not in ordered)
        if not extra:
            return ordered
        if mode == 'append':
            return ordered + extra
        raise ValueError(f"Unrecognized mode {mode!r}")




import typing

import numpy
import numpy.typing
from scipy.interpolate import interp1d

from goats.core import aliased
from goats.core import iterables
from goats.core import numerical
from goats.core import reference
from goats.core import variable


def restrict_coordinate(
    array: numpy.ndarray,
    target: float,
    axis: typing.SupportsIndex,
) -> range:
    """Restrict `array` to `target` along `axis`."""
    bounds = numerical.get_bounding_indices(array, target, axis=axis)
    if bounds.ndim == 1:
        return range(bounds[0], bounds[1]+1)
    if bounds.ndim == 2:
        lower = min(bounds[:, 0])
        upper = max(bounds[:, 1])
        return range(lower, upper+1)
    # NOTE: We can get `lower` and `upper` for N-D arrays by replacing `:` with
    # `...`. The problem is that they will be (N-1)-D arrays that we will then
    # need to split into individual index vectors before returning.
    raise NotImplementedError(f"Can't operate on {bounds.ndim}-D array")


class Restriction:
    """A subset of indices with which to create subsets of arrays."""

    def __init__(
        self,
        rule: typing.Callable[..., typing.Iterable[typing.SupportsIndex]],
        *args,
        **kwargs,
    ) -> None:
        """
        Create a new restriction.

        Parameters
        ----------
        rule : callable
            The rule by which to create a subset of array indices. This must be
            a callable object that returns an iterable of objects to be used as
            indices (e.g., `range`).

        *args
            Positional arguments to pass to `rule`.

        **kwargs
            Keyword arguments to pass to `rule`.
        """
        self.indices = rule(*args, **kwargs)

    def apply(self, array: numpy.typing.ArrayLike, axis: int=None):
        """Restrict `array` to a subset along `axis`.
        
        Parameters
        ----------
        array : array-like
            The array from which to extract a subarray corresponding to the
            restricted indices.

        axis : int, default=-1
            The axis of `array` to restrict.

        Returns
        -------
        array-like
            The subarray.
        """
        if axis is None:
            axis = -1
        permuted = numpy.moveaxis(array, axis, 0)
        result = permuted[self.indices, ...]
        return numpy.moveaxis(result, 0, axis)


def apply(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    targets: typing.Iterable[float],
    axis: int=None
) -> numpy.ndarray:
    """Interpolate `array` to target values over `coordinate`."""
    interpolated = [
        _apply_interp1d(array, reference, target, axis=axis)
        for target in targets
    ]
    if reference.ndim == 2:
        return numpy.swapaxes(interpolated, 0, 1)
    return numpy.array(interpolated)


def _apply_interp1d(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    target: float,
    axis: int=None,
) -> typing.List[float]:
    """Interpolate data to `target` along the leading axis."""
    if target in reference:
        idx = numerical.find_nearest(reference, target).index
        return array[idx]
    if axis is not None:
        restriction = Restriction(
            restrict_coordinate,
            reference,
            target,
            axis=axis,
        )
        ref = restriction.apply(reference, axis=axis)
        arr = restriction.apply(array, axis=axis)
    else:
        ref = reference
        arr = array
    # HACK: If `ref.size == 1`, the restriction process constrained the
    # reference array to a single value. There is no point in attempting to
    # interpolate over a single reference value since interpolation will fail
    # with fewer than two interpolant values. This may mean that `target` is
    # numerically outside the range of `reference`, which could lead to
    # pathological cases, but the solution would require some care to ensure
    # that `target` is far enough from the extrema of `reference` to represent a
    # truly erroneous value.
    if reference.ndim == 2:
        if ref.size == 1:
            return numpy.array([numpy.squeeze(y) for y in arr])
        interps = [interp1d(x, y, axis=0) for x, y in zip(ref, arr)]
        return numpy.array([interp(target) for interp in interps])
    if ref.size == 1:
        return numpy.squeeze(arr)
    interp = interp1d(ref, arr, axis=0)
    return interp(target)




import typing

from goats.core import aliased
from goats.core import physical
from goats.core import iterables
from goats.core import measurable
from goats.core import metadata


class Assumption(physical.Vector):
    """A measurable parameter argument.
    
    This object behaves like a vector in the sense that it is a multi-valued
    measurable quantity with a unit, but users can also cast a single-valued
    assumption to the built-in `int` and `float` types.
    """

    def __getitem__(self, arg):
        if isinstance(arg, str):
            return super().__getitem__(arg)
        values = super().__getitem__(arg)
        try:
            iter(values)
        except TypeError:
            result = physical.Scalar(values, unit=self.unit, name=self.name)
        else:
            result = [
                physical.Scalar(value, unit=self.unit, name=self.name)
                for value in values
            ]
        return result

    def __float__(self):
        """Represent a single-valued measurement as a `float`."""
        return self._cast_to(float)

    def __int__(self):
        """Represent a single-valued measurement as a `int`."""
        return self._cast_to(int)

    Numeric = typing.TypeVar('Numeric', int, float)

    def _cast_to(self, __type: typing.Type[Numeric]) -> Numeric:
        """Internal method for casting to numeric type."""
        nv = len(self.data)
        if nv == 1:
            return __type(self.data[0])
        errmsg = f"Can't convert measurement with {nv!r} values to {__type}"
        raise TypeError(errmsg) from None


class Option(metadata.NameMixin, iterables.ReprStrMixin):
    """An unmeasurable parameter argument."""

    def __init__(
        self,
        __value,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None:
        self.value = __value
        """The value of this optional parameter."""
        self._name = aliased.MappingKey(name or '')
        self.display.register('value')
        self.display['__str__'] = "{value}"
        self.display['__repr__'] = "{value}"
        self.display['__repr__'].separator = ', '
        if self._name:
            self.display.register('name')
            self.display['__str__'].insert(0, "'{name}': ")
            self.display['__repr__'].insert(1, "'{name}'")

    def __eq__(self, other):
        """True if `other` is equivalent to this option's value."""
        if isinstance(other, Option):
            return other.value == self.value
        return other == self.value


class Interface(aliased.Mapping):
    """An interface to operational assumptions and options."""

    def __getitem__(self, __k: str):
        """Create the appropriate object for the named parameter."""
        try:
            this = super().__getitem__(__k)
        except KeyError:
            raise KeyError(f"No parameter corresponding to {__k!r}") from None
        value = this['value']
        aliases = self.alias(__k, include=True)
        if unit := this['unit']:
            return Assumption(value, unit=unit, name=aliases)
        return Option(value, name=aliases)





import typing

import numpy

from goats.core import aliased
from goats.core import constant
from goats.core import index
from goats.core import iterables
from goats.core import metric
from goats.core import physical
from goats.core import variable


class Context:
    """The context of an observation."""

    def __init__(
        self,
        indices: typing.Mapping[str, index.Quantity],
        scalars: typing.Mapping[str, physical.Scalar]=None
    ) -> None:
        self._indices = indices
        self._scalars = scalars
        self._assumptions = None
        self._axes = None

    @property
    def axes(self):
        """The arrays of axis values relevant to the observation."""
        if self._axes is None:
            items = (
                self._indices.items(aliased=True)
                if isinstance(self._indices, aliased.Mapping)
                else self._indices.items()
            )
            axes = {
                k: physical.Array(
                    index.values,
                    unit=index.unit,
                    name=index.name,
                )
                for k, index in items
                # HACK: This handles cases in which a function added a
                # non-standard axis. For example, integral flux removes the
                # 'energy' axis and adds a 'minimum energy' axis. A long-term
                # solution should treat non-standard axes as equivalent to
                # standard axes.
                if index
            }
            self._axes = aliased.Mapping(axes)
        return self._axes

    @property
    def assumptions(self):
        """The physical assumptions relevant to this observation."""
        if self._assumptions is None:
            items = (
                self._scalars.items(aliased=True)
                if isinstance(self._scalars, aliased.Mapping)
                else self._scalars.items()
            )
            assumptions = {
                k: constant.Assumption(scalar)
                for k, scalar in items if scalar
            }
            self._assumptions = aliased.Mapping(assumptions)
        return self._assumptions

    def __eq__(self, __o) -> bool:
        """True if two contexts have equal axes and assumptions."""
        if isinstance(__o, Context):
            return all(
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('axes', 'assumptions')
            )
        return NotImplemented


class Quantity(iterables.ReprStrMixin):
    """The result of observing an observable quantity."""

    def __init__(
        self,
        __v: variable.Quantity,
        context: Context,
    ) -> None:
        self._quantity = __v
        self._context = context
        self._array = None
        self._data = None
        self._parameters = None

    @property
    def array(self):
        """The observed data array, with singular dimensions removed.
        
        This property primarily provides a shortcut for cases in which the
        result of observing an N-dimensional quantity is an effectively
        M-dimensional quantity, with M<N, but singular dimensions cause it to
        appear to have higher dimensionality.
        """
        if self._array is None:
            self._array = numpy.array(self.data).squeeze()
        return self._array

    @property
    def data(self):
        """The observed variable quantity.
        
        This property provides direct access to the variable-quantity interface,
        as well as to metadata properties of the observed quantity.
        """
        if self._data is None:
            self._data = self._quantity
        return self._data

    @property
    def parameters(self):
        """The names of scalar assumptions relevant to this observation."""
        if self._parameters is None:
            self._parameters = list(self._context.assumptions)
        return self._parameters

    def __getitem__(self, __x):
        """Get context items or update the unit.
        
        Parameters
        ----------
        __x : string
            If `__x` names a known array axis or physical assumption, return
            that quantity. If `__x` is a valid unit for this observed quantity,
            return a new instance with updated unit.
        """
        if not isinstance(__x, (str, aliased.MappingKey, metric.Unit)):
            raise TypeError(
                f"{__x!r} must name a context item or a unit."
                "Use the array property to access data values."
            ) from None
        if __x in self._context.axes:
            return self._context.axes[__x]
        if __x in self._context.assumptions:
            return self._context.assumptions[__x]
        return type(self)(self.data[__x], self._context)

    def __eq__(self, __o) -> bool:
        """True if two instances have equivalent attributes."""
        if isinstance(__o, Quantity):
            return all(
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('data', 'context')
            )
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        attrs = [
            f"unit='{self.data.unit}'",
            f"dimensions={self.data.axes}",
            f"parameters={self.parameters}",
        ]
        return f"'{self.data.name}': {', '.join(attrs)}"




"""Tools for managing variable quantities in datasets."""

import numbers
import typing

import numpy
import numpy.typing

from goats.core import aliased
from goats.core import datafile
from goats.core import metric
from goats.core import metadata
from goats.core import reference
from goats.core import physical


Instance = typing.TypeVar('Instance', bound='Quantity')


class Quantity(physical.Array, metadata.AxesMixin):
    """A class representing a dataset variable."""

    @typing.overload
    def __init__(
        self: Instance,
        __data: numpy.typing.ArrayLike,
        *,
        axes: typing.Iterable[str],
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None:
        """Create a new variable from scratch."""

    @typing.overload
    def __init__(
        self: Instance,
        __data: physical.Array,
        *,
        axes: typing.Iterable[str],
    ) -> None:
        """Create a new variable from an array."""

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> None:
        """Create a new variable from an existing variable."""

    def __init__(self, __data, **meta) -> None:
        super().__init__(__data, **meta)
        parsed = self.parse_attrs(__data, meta, axes=())
        axes = (
            metadata.Axes(parsed['axes'])
            or [f'x{i}' for i in range(self.ndim)]
        )
        self._axes = tuple(axes)
        self.meta.register('axes')
        self.naxes = len(self.axes)
        """The number of indexable axes in this variable's array."""
        if self.naxes != self.ndim:
            raise ValueError(
                f"Number of axes ({self.naxes})"
                f" must equal number of array dimensions ({self.ndim})"
            )
        self.display.register('axes')
        self.display['__str__'].append("axes={axes}")
        self.display['__repr__'].append("axes={axes}")

    def parse_attrs(self, this, meta: dict, **targets):
        if isinstance(this, physical.Array) and not isinstance(this, Quantity):
            attrs = {'unit', 'name'}
            meta.update({k: getattr(this, k) for k in attrs})
            this = this.data
        return super().parse_attrs(this, meta, **targets)

    def _ufunc_hook(self, ufunc, *inputs):
        """Convert input arrays into arrays appropriate to `ufunc`."""
        multiplicative = ufunc.__name__ in {'multiply', 'divide', 'true_divide'}
        correct_type = all(isinstance(v, type(self)) for v in inputs)
        if multiplicative and correct_type:
            axes = self.axes.merge(*[v.axes for v in inputs])
            tmp = {}
            for v in inputs:
                tmp.update(v.shape_dict)
            full_shape = tuple(tmp[d] for d in axes)
            indices = numpy.ix_(*[range(i) for i in full_shape])
            arrays = []
            for v in inputs:
                idx = tuple(indices[axes.index(d)] for d in v.shape_dict)
                arrays.append(v._get_array(idx))
            return arrays
        return super()._ufunc_hook(ufunc, *inputs)

    def array_contains(self, value: numbers.Real):
        """True if `value` is in this variable quantity's array.

        Parameters
        ----------
        value : real
            The value to check for among this variable's values.
        
        Notes
        -----
        This method exists to handle cases in which floating-point arithmetic
        has caused a numeric operation to return an imprecise result, especially
        for small numbers (e.g., converting energy from eV to J). It will first
        check the built-in `__contains__` method via `in` before attempting to
        determine if `value` is close enough to count, albeit within a very
        strict tolerance.
        """
        if value in self._array:
            return True
        if value < numpy.min(self._array) or value > numpy.max(self._array):
            return False
        return numpy.any([numpy.isclose(value, self._array, atol=0.0)])

    @typing.overload
    def __getitem__(
        self: Instance,
        unit: typing.Union[str, metric.Unit],
    ) -> Instance: ...

    @typing.overload
    def __getitem__(
        self,
        *args: physical.IndexLike,
    ) -> typing.Union[physical.Scalar, physical.Array, Instance]: ...

    def __getitem__(self, *args):
        result = super().__getitem__(*args)
        if isinstance(result, physical.Array) and result.ndim == self.axes:
            return Quantity(result, axes=self.axes)
        return result

    def __eq__(self, other: typing.Any):
        return (
            super().__eq__(other) if getattr(other, 'axes', None) == self.axes
            else False
        )

    @property
    def shape_dict(self) -> typing.Dict[str, int]:
        """Label and size for each axis."""
        return dict(zip(self.axes, self.shape))


@Quantity.implements(numpy.squeeze)
def _squeeze(v: Quantity, **kwargs):
    """Remove singular axes."""
    data = v._array.squeeze(**kwargs)
    axes = tuple(
        a for a, d in zip(v.axes, v.shape)
        if d != 1
    )
    return Quantity(data, unit=v.unit, name=v.name, axes=axes)


@Quantity.implements(numpy.mean)
def _mean(v: Quantity, **kwargs):
    """Compute the mean of the underlying array."""
    data = v._array.mean(**kwargs)
    axis = kwargs.get('axis')
    if axis is None:
        return data
    axes = tuple(a for a in v.axes if v.axes.index(a) != axis)
    name = [f"mean({name})" for name in v.name]
    return Quantity(data, unit=v.unit, name=name, axes=axes)


S = typing.TypeVar('S', bound='Interface')


class Interface(aliased.Mapping):
    """An interface to dataset variables.
    
    This class provides aliased key-based access to all variables in a dataset.
    It converts each requested dataset variable into a `~variable.Quantity`
    instance with the appropriate for its metric system.
    """

    def __init__(
        self,
        dataset: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        super().__init__(dataset.variables)
        self._system = metric.System(system or 'mks')
        self._cache = {}

    @property
    def system(self):
        """This observer's metric system."""
        return self._system

    def __getitem__(self, key: str) -> Quantity:
        """Create the named variable, if possible."""
        if key in self._cache:
            return self._cache[key]
        datavar = super().__getitem__(key)
        variable = Quantity(
            datavar.data,
            unit=standardize(datavar.unit),
            axes=datavar.axes,
            name=reference.ALIASES.get(key, key),
        )
        converted = (
            # HACK: This special case comes from the fact that EPREM outputs
            # mass in 'nucleon', which is really a mass number. The `eprem`
            # subpackage should ultimately handle this logic (and maybe the
            # other special cases in `substitutions`)
            variable[self.system['mass'].unit]
            if key in {'mass', 'm'}
            else variable[str(self.system)]
        )
        self._cache[key] = converted
        return converted


substitutions = {
    'julian date': 'day',
    'shell': '1',
    'cos(mu)': '1',
    'e-': 'e',
    '# / cm^2 s sr MeV': '# / cm^2 s sr MeV/nuc',
}
"""Conversions from non-standard units."""

T = typing.TypeVar('T')

def standardize(unit: T):
    """Replace this unit string with a standard unit string, if possible.

    This function looks for `unit` in the known conversions and returns the
    standard unit string if it exists. If this doesn't find a standard unit
    string, it just returns the input.
    """
    unit = substitutions.get(str(unit), unit)
    if '/' in unit:
        num, den = str(unit).split('/', maxsplit=1)
        unit = ' / '.join((num.strip(), f"({den.strip()})"))
    return unit





import typing

from goats.core import aliased
from goats.core import metadata
from goats.core import metric
from goats.core import observable
from goats.core import observing
from goats.core import reference
from goats.core import spelling
from goats.core import symbolic


class Interface:
    """The base class for all observers."""

    def __init__(
        self,
        __data: observing.Dataset,
        *unobservable: str,
        system: str='mks',
        apply: typing.Type[observing.Application]=None,
    ) -> None:
        """Initialize this instance.
        
        Parameters
        ----------
        __data : `~observing.Dataset`
            The observer's dataset.

        *unobservable : string
            The names of variable quantities from this observer's dataset to
            exclude from the set of formally observable quantities. These
            quantities will still be accessible as variable quantities.

        system : string, default='mks'
            The metric system to use for variable and observable quantities.

        apply : type of observing application, optional
            A subclass of `~observing.Application` with which to evaluate
            operational parameters and variable quantities during the observing
            process. If `application` is absent, this class will use an instance
            of the base class.
        """
        self._data = __data
        self._unobservable = unobservable
        self._system = metric.System(system)
        self._application_type = apply or observing.Application
        self._quantities = None
        self._application = None
        self._spellcheck = None

    def reset(self, source=None, **kwargs):
        """Reset the dataset interface.
        
        The base implementation resets various data-related attributes to their
        uninitialized values and sets `source`, if given, as the new target from
        which to read data. Concrete subclasses may wish to implement additional
        data-related logic (e.g., reinitialize interfaces), possibly after
        modifying `source` (e.g., to normalize a path).

        Parameters
        ----------
        source, optional
            The new source of this observer's data. The acceptable type(s) will
            be observer-specific.
        """
        self._quantities = None
        self._application = None
        self._spellcheck = None
        if source is not None:
            self._data.readfrom(source, **kwargs)
        return self

    @property
    def system(self):
        """This observer's metric system."""
        return self._system

    @property
    def application(self):
        """This observer's observing application."""
        if self._application is None:
            self._application = self._application_type(self.quantities)
        return self._application

    @property
    def parameters(self):
        """The names of operational arguments relevant to this observer."""
        keys = self.quantities.constants.keys(aliased=True)
        return aliased.KeyMap(*keys)

    @property
    def observables(self):
        """The names of formally observable quantities.
        
        This property contains the names of all primary and derived observable
        quantities. See `~observing.Interface.primary` and
        `~observing.Interface.derived` for their respective definitions. Names
        are listed as groups of aliases for each observable quantity (e.g., 'mfp
        | mean free path | mean_free_path'); any of the listed aliases is a
        valid key for that quantity.

        Note that it is also possible to symbolically compose new observable
        quantities from those listed here (e.g., 'mfp / Vr'). Therefore, this
        collection represents the minimal set of quantities that this observer
        can observe.
        """
        primary = self.quantities.primary
        derived = self.quantities.derived
        available = aliased.KeyMap(*primary, *derived)
        return available.without(*self._unobservable)

    @property
    def quantities(self):
        """An interface to this observer's physical quantities.
        
        This property represents the variable, axis-indexing, and constant
        quantities that are available from this observer's dataset. It
        incorporates this observer's metric system, and exposes objects that
        support arithmetic operations and conversion to numpy arrays.
        """
        if self._quantities is None:
            self._quantities = observing.Quantities(
                self._data,
                system=self.system,
            )
        return self._quantities

    def __getitem__(self, key: str):
        """Access an observable quantity by keyword, if possible."""
        if q := self._get_quantity(key):
            return q
        self._check_spelling(key) # -> None if `key` is spelled correctly
        raise KeyError(f"No observable for {key!r}") from None

    def _get_quantity(self, key: str):
        """Retrieve the named quantity from an interface, if possible."""
        if self.knows(key):
            name = metadata.Name(
                key if observable.iscomposed(key)
                else reference.ALIASES[key]
            )
            return observable.Quantity(
                name,
                observing.Implementation(key, self.quantities),
                application=self.application,
            )
        if key in self.quantities.variables:
            return self.quantities.variables[key]
        if key in self.quantities.constants:
            return self.quantities.constants[key]

    def knows(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> bool:
        """True if this interface can observe the named quantity."""
        if isinstance(name, str):
            return self._knows(name)
        return next((self._knows(key) for key in name), False)

    def _knows(self, key: str):
        """Internal helper for `~Interface.knows`."""
        if key in list(self.observables):
            return True
        if observable.iscomposed(key):
            expression = symbolic.Expression(key)
            return all(self._knows(term.base) for term in expression)
        return False

    def _check_spelling(self, key: str):
        """Catch misspelled names of observable quantities, if possible."""
        keys = list(self.observables) + list(self.parameters)
        if self._spellcheck is None:
            self._spellcheck = spelling.SpellChecker(*keys)
        else:
            self._spellcheck.words |= keys
        return self._spellcheck.check(key)

